# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# The code is modified from DesignPlanar to implement flipchip-compatible QDesign class. (SK/Chalmers/20210610)
"""Module containing Basic Qiskit Metal FlipChip design for CPW type
geometry."""

# from typing import TYPE_CHECKING
from typing import Tuple

# from typing import Dict as Dict_
# from typing import List, Union

from .design_base import QDesign, Dict

__all__ = ["DesignFlipChip"]


class DesignFlipChip(QDesign):
    """Metal class for a flipchip design, consisting of two chips with their device sides facing each other.
    Typically assumed to have some CPW geometries.

    Inherits QDesign class.
    """

    def __init__(
        self,
        metadata: dict = None,
        overwrite_enabled: bool = False,
        enable_renderers: bool = True,
    ):
        """Pass metadata to QDesign.

        Args:
            metadata (dict, optional): Pass to QDesign. Defaults to {}.
            overwrite_enabled (bool, optional): Passed to QDesign base class. Defaults to False.
            enable_renderers (bool, optional): Passed to QDesign base class. Defaults to True.
        """

        super().__init__(
            metadata=metadata,
            overwrite_enabled=overwrite_enabled,
            enable_renderers=enable_renderers,
        )
        self.variables[
            "sample_holder_top"] = "890um"  # how tall is the vacuum above z=0
        self.variables[
            "sample_holder_bottom"] = "1650um"  # how tall is the vacuum below z=0
        self.add_chip_info()

    def add_chip_info(self):
        """TODO How to get the values into self.chip. Will need to set up
        parser for "self.p" for design base. For now, just hard code in
        something.

        # GDSTK is using numbers based on 1 meter unit.
        # When the gds file is exported, data is converted to "user-selected" units.
        # centered at (0,0) and 9 by 6 size.

        NOTE: self._chips dict comes from QDesign base class.
        """
        # Reinitilise self._chips, then add two chips
        self._chips = {}
        self._chips["C_chip"] = Dict(material="silicon",
                                     layer_start="0",
                                     layer_end="2048")
        self._chips["Q_chip"] = Dict(material="silicon",
                                     layer_start="0",
                                     layer_end="2048")

        self._chips["C_chip"]["size"] = Dict(
            center_x="0.0mm",
            center_y="0.0mm",
            center_z="0.0mm",
            size_x="9mm",
            size_y="9mm",
            size_z="-280um",  # chip extends in negative z direction by 280 um
            sample_holder_top=self.variables["sample_holder_top"],
            sample_holder_bottom=self.variables["sample_holder_bottom"],
        )

        self._chips["Q_chip"]["size"] = Dict(
            center_x="0.0mm",
            center_y="0.0mm",
            center_z="20 um",
            size_x="9mm",
            size_y="9mm",
            size_z="280um",  # chip extends in positive z direction by 280 um
            sample_holder_top=self.variables["sample_holder_top"],
            sample_holder_bottom=self.variables["sample_holder_bottom"],
        )

    def get_x_y_for_chip(self, chip_name: str) -> Tuple[tuple, int]:
        """
        If the chip_name is in self.chips, along with entry for size
        information then return a tuple=(minx, miny, maxx, maxy). Used for
        subtraction while exporting design.

        Args:
            chip_name (str): Name of chip that you want the size of.

        Returns:
            Tuple[tuple, int]:
            tuple: The exact placement on rectangle coordinate (minx, miny, maxx, maxy).
            int: 0=all is good
            1=chip_name not in self._chips
            2=size information missing or no good
        """
        x_y_location = tuple()

        if chip_name in self._chips:
            if "size" in self._chips[chip_name]:

                size = self.parse_value(self.chips[chip_name]["size"])
                if ("center_x" in size and "center_y" in size and
                        "size_x" in size and "size_y" in size):
                    if (type(size.center_x) in [int, float] and
                            type(size.center_y) in [int, float] and
                            type(size.size_x) in [int, float] and
                            type(size.size_y) in [int, float]):
                        x_y_location = (
                            size["center_x"] - (size["size_x"] / 2.0),
                            size["center_y"] - (size["size_y"] / 2.0),
                            size["center_x"] + (size["size_x"] / 2.0),
                            size["center_y"] + (size["size_y"] / 2.0),
                        )
                        return x_y_location, 0

                    self.logger.warning(
                        f'Size information within self.chips[{chip_name}]["size"]'
                        f" is NOT an int or float.")
                    return x_y_location, 2

                self.logger.warning("center_x or center_y or size_x or size_y "
                                    f' NOT in self._chips[{chip_name}]["size"]')
                return x_y_location, 2

            self.logger.warning(
                f"Information for size in NOT in self._chips[{chip_name}]"
                ' dict. Return "None" in tuple.')
            return x_y_location, 2

        self.logger.warning(
            f'Chip name "{chip_name}" is not in self._chips dict. Return "None" in tuple.'
        )
        return x_y_location, 1



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Temp\GitHub\qiskit-metal\qiskit_metal\_gui\renderer_q3d_ui.ui'
#
# Created: Sat Jun 19 22:02:29 2021
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):

    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(651, 544)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.instructionsLabel = QtWidgets.QLabel(self.centralwidget)
        self.instructionsLabel.setObjectName("instructionsLabel")
        self.gridLayout.addWidget(self.instructionsLabel, 0, 0, 1, 1)
        self.instructionsLabel_2 = QtWidgets.QLabel(self.centralwidget)
        self.instructionsLabel_2.setAlignment(QtCore.Qt.AlignCenter)
        self.instructionsLabel_2.setObjectName("instructionsLabel_2")
        self.gridLayout.addWidget(self.instructionsLabel_2, 0, 1, 1, 1)
        self.listView = QtWidgets.QListView(self.centralwidget)
        self.listView.setObjectName("listView")
        self.gridLayout.addWidget(self.listView, 1, 0, 1, 1)
        self.treeView = QTreeView_Base(self.centralwidget)
        self.treeView.setRootIsDecorated(False)
        self.treeView.setObjectName("treeView")
        self.gridLayout.addWidget(self.treeView, 1, 1, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.refreshButton = QtWidgets.QPushButton(self.centralwidget)
        self.refreshButton.setObjectName("refreshButton")
        self.horizontalLayout_4.addWidget(self.refreshButton)
        self.selectAllButton = QtWidgets.QPushButton(self.centralwidget)
        self.selectAllButton.setObjectName("selectAllButton")
        self.horizontalLayout_4.addWidget(self.selectAllButton)
        self.deselectAllButton = QtWidgets.QPushButton(self.centralwidget)
        self.deselectAllButton.setObjectName("deselectAllButton")
        self.horizontalLayout_4.addWidget(self.deselectAllButton)
        self.gridLayout.addLayout(self.horizontalLayout_4, 2, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.confirmButton = QtWidgets.QPushButton(self.centralwidget)
        self.confirmButton.setObjectName("confirmButton")
        self.horizontalLayout.addWidget(self.confirmButton)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 1, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 651, 18))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        self.selectAllButton.clicked.connect(
            MainWindow.select_all)  # type: ignore
        self.deselectAllButton.clicked.connect(
            MainWindow.deselect_all)  # type: ignore
        self.refreshButton.clicked.connect(MainWindow.refresh)  # type: ignore
        self.confirmButton.clicked.connect(
            MainWindow.choose_checked_components)  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Q3D Renderer"))
        self.instructionsLabel.setText(
            _translate("MainWindow", "Check off components to export:"))
        self.instructionsLabel_2.setText(
            _translate("MainWindow", "Renderer options"))
        self.refreshButton.setText(_translate("MainWindow", "Refresh List"))
        self.selectAllButton.setText(_translate("MainWindow", "Select All"))
        self.deselectAllButton.setText(_translate("MainWindow", "Deselect All"))
        self.confirmButton.setText(_translate("MainWindow",
                                              "Confirm Selection"))


from . import main_window_rc_rc
from .tree_view_base import QTreeView_Base



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './widgets/create_component_window/parameter_entry_window_ui.ui',
# licensing of './widgets/create_component_window/parameter_entry_window_ui.ui' applies.
#
# Created: Sat Jun 19 22:02:30 2021
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):

    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(753, 600)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.explanatory_label = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                           QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(
            self.explanatory_label.sizePolicy().hasHeightForWidth())
        self.explanatory_label.setSizePolicy(sizePolicy)
        self.explanatory_label.setObjectName("explanatory_label")
        self.verticalLayout_2.addWidget(self.explanatory_label)
        self.widget = QtWidgets.QWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                           QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(20)
        sizePolicy.setHeightForWidth(
            self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setObjectName("widget")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.widget)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.edit_box = QtWidgets.QWidget(self.widget)
        self.edit_box.setObjectName("edit_box")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.edit_box)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.widget_3 = QtWidgets.QWidget(self.edit_box)
        self.widget_3.setObjectName("widget_3")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.widget_3)
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.add_k_v_row_button = QtWidgets.QPushButton(self.widget_3)
        self.add_k_v_row_button.setObjectName("add_k_v_row_button")
        self.horizontalLayout_5.addWidget(self.add_k_v_row_button)
        self.nest_dictionary_button = QtWidgets.QPushButton(self.widget_3)
        self.nest_dictionary_button.setObjectName("nest_dictionary_button")
        self.horizontalLayout_5.addWidget(self.nest_dictionary_button)
        self.remove_button = QtWidgets.QPushButton(self.widget_3)
        self.remove_button.setObjectName("remove_button")
        self.horizontalLayout_5.addWidget(self.remove_button)
        self.verticalLayout_4.addWidget(self.widget_3)
        self.qcomponent_param_tree_view = TreeViewParamEntry(self.edit_box)
        self.qcomponent_param_tree_view.setObjectName(
            "qcomponent_param_tree_view")
        self.verticalLayout_4.addWidget(self.qcomponent_param_tree_view)
        self.parameter_entry_form = QtWidgets.QWidget(self.edit_box)
        self.parameter_entry_form.setObjectName("parameter_entry_form")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.parameter_entry_form)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.verticalLayout_4.addWidget(self.parameter_entry_form)
        self.create_qcomp_button = QtWidgets.QPushButton(self.edit_box)
        self.create_qcomp_button.setObjectName("create_qcomp_button")
        self.verticalLayout_4.addWidget(self.create_qcomp_button)
        self.horizontalLayout_4.addWidget(self.edit_box)
        self.help_box = QtWidgets.QWidget(self.widget)
        self.help_box.setObjectName("help_box")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.help_box)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.help_button_holder = QtWidgets.QWidget(self.help_box)
        self.help_button_holder.setObjectName("help_button_holder")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.help_button_holder)
        self.horizontalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.verticalLayout_5.addWidget(self.help_button_holder)
        self.help_tab_tabWidget = QtWidgets.QTabWidget(self.help_box)
        self.help_tab_tabWidget.setObjectName("help_tab_tabWidget")
        self.tab_source = QtWidgets.QWidget()
        self.tab_source.setObjectName("tab_source")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.tab_source)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.help_tab_tabWidget.addTab(self.tab_source, "")
        self.tab_help = QtWidgets.QWidget()
        self.tab_help.setObjectName("tab_help")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.tab_help)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.help_tab_tabWidget.addTab(self.tab_help, "")
        self.verticalLayout_5.addWidget(self.help_tab_tabWidget)
        self.horizontalLayout_4.addWidget(self.help_box)
        self.verticalLayout_2.addWidget(self.widget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        self.help_tab_tabWidget.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(
            QtWidgets.QApplication.translate("MainWindow", "MainWindow", None,
                                             -1))
        self.explanatory_label.setText(
            QtWidgets.QApplication.translate("MainWindow",
                                             "Options to create with", None,
                                             -1))
        self.add_k_v_row_button.setText(
            QtWidgets.QApplication.translate("MainWindow", "Add Row", None, -1))
        self.nest_dictionary_button.setText(
            QtWidgets.QApplication.translate("MainWindow", "Nest Dictionary",
                                             None, -1))
        self.remove_button.setText(
            QtWidgets.QApplication.translate("MainWindow", "Remove", None, -1))
        self.create_qcomp_button.setText(
            QtWidgets.QApplication.translate("MainWindow", "Create", None, -1))
        self.help_tab_tabWidget.setTabText(
            self.help_tab_tabWidget.indexOf(self.tab_source),
            QtWidgets.QApplication.translate("MainWindow", "Source", None, -1))
        self.help_tab_tabWidget.setTabText(
            self.help_tab_tabWidget.indexOf(self.tab_help),
            QtWidgets.QApplication.translate("MainWindow", "Help", None, -1))


from .model_view.tree_view_param_entry import TreeViewParamEntry



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=broad-except
# pylint: disable-msg=too-many-public-methods
# pylint: disable-msg=import-error
# pylint: disable-msg=protected-access
"""Qiskit Metal unit tests analyses functionality."""

import unittest
from unittest.mock import MagicMock
import matplotlib.pyplot as _plt

from qiskit_metal import designs
from qiskit_metal.renderers import setup_default
from qiskit_metal.renderers.renderer_ansys.ansys_renderer import QAnsysRenderer
from qiskit_metal.renderers.renderer_ansys.q3d_renderer import QQ3DRenderer
from qiskit_metal.renderers.renderer_ansys.hfss_renderer import QHFSSRenderer
from qiskit_metal.renderers.renderer_base.renderer_base import QRenderer
from qiskit_metal.renderers.renderer_base.renderer_gui_base import QRendererGui
from qiskit_metal.renderers.renderer_gds.gds_renderer import QGDSRenderer
from qiskit_metal.renderers.renderer_mpl.mpl_interaction import MplInteraction
from qiskit_metal.renderers.renderer_gmsh.gmsh_renderer import QGmshRenderer
from qiskit_metal.renderers.renderer_elmer.elmer_renderer import QElmerRenderer
from qiskit_metal.renderers.renderer_ansys_pyaedt.hfss_renderer_eigenmode_aedt import QHFSSEigenmodePyaedt
from qiskit_metal.renderers.renderer_ansys_pyaedt.hfss_renderer_drivenmodal_aedt import QHFSSDrivenmodalPyaedt
from qiskit_metal.renderers.renderer_ansys_pyaedt.q3d_renderer_aedt import QQ3DPyaedt

from qiskit_metal.renderers.renderer_ansys import ansys_renderer

from qiskit_metal.qgeometries.qgeometries_handler import QGeometryTables
from qiskit_metal.qlibrary.qubits.transmon_pocket import TransmonPocket
from qiskit_metal import draw


class TestRenderers(unittest.TestCase):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_renderer_instanitate_qansys_renderer(self):
        """Test instantiation of QAnsysRenderer in ansys_renderer.py"""
        design = designs.DesignPlanar()
        try:
            QAnsysRenderer(design, initiate=False)
        except Exception:
            self.fail("QAnsysRenderer() failed")

    def test_renderer_instantiate_gdsrender(self):
        """Test instantiation of QGDSRenderer in gds_renderer.py."""
        design = designs.DesignPlanar()
        try:
            QGDSRenderer(design)
        except Exception:
            self.fail("QGDSRenderer(design) failed")

        try:
            QGDSRenderer(design, initiate=False)
        except Exception:
            self.fail("QGDSRenderer(design, initiate=False) failed")

        try:
            QGDSRenderer(design, initiate=False, render_template={})
        except Exception:
            self.fail(
                "QGDSRenderer(design, initiate=False, render_template={}) failed"
            )

        try:
            QGDSRenderer(design, initiate=False, render_options={})
        except Exception:
            self.fail(
                "QGDSRenderer(design, initiate=False, render_options={}) failed"
            )

    def test_renderer_instantiate_mplinteraction(self):
        """Test instantiation of MplInteraction in mpl_interaction.py."""
        try:
            MplInteraction(_plt)
        except Exception:
            self.fail("MplInteraction(None) failed")

    def test_renderer_instantiate_qq3d_renderer(self):
        """Test instantiation of QQ3DRenderer in q3d_renderer.py."""
        design = designs.DesignPlanar()
        try:
            QQ3DRenderer(design, initiate=False)
        except Exception:
            self.fail("QQ3DRenderer failed")

    def test_renderer_instantiate_qhfss_renderer(self):
        """Test instantiation of QHFSSRenderer in hfss_renderer.py."""
        design = designs.DesignPlanar()
        try:
            QHFSSRenderer(design, initiate=False)
        except Exception:
            self.fail("QHFSSRenderer failed")

    def test_renderer_instantiate_qgmsh_renderer(self):
        """Test instantiation of QGmshRenderer in gmsh_renderer.py."""
        design = designs.DesignPlanar()
        try:
            QGmshRenderer(design)
        except Exception:
            self.fail("QGmshRenderer(design) failed")

        try:
            QGmshRenderer(design, initiate=False)
        except Exception:
            self.fail(
                "QGmshRenderer(design, initiate=False) failed in DesignPLanar")

        try:
            QGmshRenderer(design, initiate=False, options={})
        except Exception:
            self.fail(
                "QGmshRenderer(design, initiate=False, options={}) failed in DesignPLanar"
            )

    def test_renderer_instantiate_qelmer_renderer(self):
        """Test instantiation of QElmerRenderer in elmer_renderer.py."""
        design = designs.MultiPlanar()
        try:
            QElmerRenderer(design)
        except Exception:
            self.fail("QElmerRenderer(design) failed")

        try:
            QElmerRenderer(design, initiate=False)
        except Exception:
            self.fail(
                "QElmerRenderer(design, initiate=False) failed in MultiPlanar")

        try:
            QElmerRenderer(design, initiate=False, options={})
        except Exception:
            self.fail(
                "QElmerRenderer(design, initiate=False, options={}) failed in MultiPlanar"
            )

    def test_renderer_qansys_renderer_options(self):
        """Test that defaults in QAnsysRenderer were not accidentally changed."""
        design = designs.DesignPlanar()
        renderer = QAnsysRenderer(design, initiate=False)
        options = renderer.default_options

        self.assertEqual(len(options), 14)
        self.assertEqual(options['Lj'], '10nH')
        self.assertEqual(options['Cj'], 0)
        self.assertEqual(options['_Rj'], 0)
        self.assertEqual(options['max_mesh_length_jj'], '7um')
        self.assertEqual(options['max_mesh_length_port'], '7um')
        self.assertEqual(options['project_path'], None)
        self.assertEqual(options['project_name'], None)
        self.assertEqual(options['design_name'], None)
        self.assertEqual(options['x_buffer_width_mm'], 0.2)
        self.assertEqual(options['y_buffer_width_mm'], 0.2)
        self.assertEqual(options['wb_threshold'], '400um')
        self.assertEqual(options['wb_offset'], '0um')
        self.assertEqual(options['wb_size'], 5)

        self.assertEqual(len(options['plot_ansys_fields_options']), 13)
        self.assertEqual(options['plot_ansys_fields_options']['name'],
                         "NAME:Mag_E1")
        self.assertEqual(
            options['plot_ansys_fields_options']['UserSpecifyName'], '0')
        self.assertEqual(
            options['plot_ansys_fields_options']['UserSpecifyFolder'], '0')
        self.assertEqual(options['plot_ansys_fields_options']['QuantityName'],
                         "Mag_E")
        self.assertEqual(options['plot_ansys_fields_options']['PlotFolder'],
                         "E Field")
        self.assertEqual(options['plot_ansys_fields_options']['StreamlinePlot'],
                         "False")
        self.assertEqual(
            options['plot_ansys_fields_options']['AdjacentSidePlot'], "False")
        self.assertEqual(options['plot_ansys_fields_options']['FullModelPlot'],
                         "False")
        self.assertEqual(options['plot_ansys_fields_options']['IntrinsicVar'],
                         "Phase=\'0deg\'")
        self.assertEqual(options['plot_ansys_fields_options']['PlotGeomInfo_0'],
                         "1")
        self.assertEqual(options['plot_ansys_fields_options']['PlotGeomInfo_1'],
                         "Surface")
        self.assertEqual(options['plot_ansys_fields_options']['PlotGeomInfo_2'],
                         "FacesList")
        self.assertEqual(options['plot_ansys_fields_options']['PlotGeomInfo_3'],
                         "1")

    def test_renderer_qansysrenderer_default_setup(self):
        """Test that default_setup in QAnsysRenderer have not been accidentally changed."""
        default_setup = QAnsysRenderer.default_setup

        self.assertEqual(len(default_setup), 4)
        self.assertEqual(len(default_setup['drivenmodal']), 8)
        self.assertEqual(len(default_setup['eigenmode']), 9)
        self.assertEqual(len(default_setup['q3d']), 12)
        self.assertEqual(default_setup['port_inductor_gap'], '10um')

        self.assertEqual(default_setup['drivenmodal']['name'], "Setup")
        self.assertEqual(default_setup['drivenmodal']['freq_ghz'], '5.0')
        self.assertEqual(default_setup['drivenmodal']['max_delta_s'], '0.1')
        self.assertEqual(default_setup['drivenmodal']['max_passes'], '10')
        self.assertEqual(default_setup['drivenmodal']['min_passes'], '1')
        self.assertEqual(default_setup['drivenmodal']['min_converged'], '1')
        self.assertEqual(default_setup['drivenmodal']['pct_refinement'], '30')
        self.assertEqual(default_setup['drivenmodal']['basis_order'], '1')

        self.assertEqual(default_setup['eigenmode']['name'], "Setup")
        self.assertEqual(default_setup['eigenmode']['min_freq_ghz'], '1')
        self.assertEqual(default_setup['eigenmode']['n_modes'], '1')
        self.assertEqual(default_setup['eigenmode']['max_delta_f'], '0.5')
        self.assertEqual(default_setup['eigenmode']['max_passes'], '10')
        self.assertEqual(default_setup['eigenmode']['min_passes'], '1')
        self.assertEqual(default_setup['eigenmode']['min_converged'], '1')
        self.assertEqual(default_setup['eigenmode']['pct_refinement'], '30')
        self.assertEqual(default_setup['eigenmode']['basis_order'], '-1')

        self.assertEqual(default_setup['q3d']['name'], 'Setup')
        self.assertEqual(default_setup['q3d']['freq_ghz'], '5.0')
        self.assertEqual(default_setup['q3d']['save_fields'], 'False')
        self.assertEqual(default_setup['q3d']['enabled'], 'True')
        self.assertEqual(default_setup['q3d']['max_passes'], '15')
        self.assertEqual(default_setup['q3d']['min_passes'], '2')
        self.assertEqual(default_setup['q3d']['min_converged_passes'], '2')
        self.assertEqual(default_setup['q3d']['percent_error'], '0.5')
        self.assertEqual(default_setup['q3d']['percent_refinement'], '30')
        self.assertEqual(default_setup['q3d']['auto_increase_solution_order'],
                         'True')
        self.assertEqual(default_setup['q3d']['solution_order'], 'High')
        self.assertEqual(default_setup['q3d']['solver_type'], 'Iterative')

    def test_renderer_qelmer_renderer_default_setup(self):
        """Test that default_setup in QElmerRenderer have not been accidentally changed."""
        default_setup = QElmerRenderer.default_setup

        self.assertEqual(len(default_setup), 4)
        self.assertEqual(len(default_setup['capacitance']), 12)
        self.assertEqual(len(default_setup['eigenmode']), 0)
        self.assertEqual(len(default_setup['materials']), 2)
        self.assertEqual(len(default_setup['constants']), 2)

        self.assertEqual(
            default_setup['capacitance']['Calculate_Electric_Field'], True)
        self.assertEqual(
            default_setup['capacitance']['Calculate_Electric_Energy'], True)
        self.assertEqual(
            default_setup['capacitance']['Calculate_Capacitance_Matrix'], True)
        self.assertEqual(
            default_setup['capacitance']['Capacitance_Matrix_Filename'],
            'cap_matrix.txt')
        self.assertEqual(default_setup['capacitance']['Linear_System_Solver'],
                         'Iterative')
        self.assertEqual(
            default_setup['capacitance']['Steady_State_Convergence_Tolerance'],
            1e-05)
        self.assertEqual(
            default_setup['capacitance']
            ['Nonlinear_System_Convergence_Tolerance'], 1e-07)
        self.assertEqual(
            default_setup['capacitance']['Nonlinear_System_Max_Iterations'], 20)
        self.assertEqual(
            default_setup['capacitance']['Linear_System_Convergence_Tolerance'],
            1e-10)
        self.assertEqual(
            default_setup['capacitance']['Linear_System_Max_Iterations'], 500)
        self.assertEqual(
            default_setup['capacitance']['Linear_System_Iterative_Method'],
            'BiCGStab')
        self.assertEqual(
            default_setup['capacitance']['BiCGstabl_polynomial_degree'], 2)

        self.assertEqual(default_setup['materials'], ['vacuum', 'silicon'])

        self.assertEqual(default_setup['constants']['Permittivity_of_Vacuum'],
                         8.8542e-12)
        self.assertEqual(default_setup['constants']['Unit_Charge'], 1.602e-19)

    def test_renderer_qq3d_render_options(self):
        """Test that defaults in QQ3DRenderer were not accidentally changed."""
        design = designs.DesignPlanar()
        renderer = QQ3DRenderer(design, initiate=False)
        options = renderer.q3d_options

        self.assertEqual(renderer.name, 'q3d')

        self.assertEqual(len(options), 2)
        self.assertEqual(options['material_type'], 'pec')
        self.assertEqual(options['material_thickness'], '200nm')

    def test_renderer_hfss_render_options(self):
        """Test that defaults in QHFSSRender were not accidentally changed."""
        design = designs.DesignPlanar()
        renderer = QHFSSRenderer(design, initiate=False)
        options = renderer.hfss_options

        self.assertEqual(renderer.name, 'hfss')
        self.assertEqual(len(options), 1)
        self.assertEqual(options['port_inductor_gap'], '10um')

    def test_renderer_gdsrenderer_options(self):
        """Test that default_options in QGDSRenderer were not accidentally
        changed."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)
        options = renderer.default_options

        self.assertEqual(len(options), 17)
        self.assertEqual(options['short_segments_to_not_fillet'], 'True')
        self.assertEqual(options['check_short_segments_by_scaling_fillet'],
                         '2.0')
        self.assertEqual(options['gds_unit'], '1')
        self.assertEqual(options['ground_plane'], 'True')
        self.assertEqual(options['negative_mask']['main'], [])
        self.assertEqual(options['corners'], 'natural')
        self.assertEqual(options['tolerance'], '0.00001')
        self.assertEqual(options['precision'], '0.000000001')
        self.assertEqual(options['width_LineString'], '10um')
        self.assertEqual(options['path_filename'],
                         '../resources/Fake_Junctions.GDS')
        self.assertEqual(options['junction_pad_overlap'], '5um')
        self.assertEqual(options['max_points'], '199')
        self.assertEqual(options['bounding_box_scale_x'], '1.2')
        self.assertEqual(options['bounding_box_scale_y'], '1.2')

        self.assertEqual(options['fabricate'], 'False')

        self.assertEqual(len(options['cheese']), 9)
        self.assertEqual(len(options['no_cheese']), 5)

        self.assertEqual(options['cheese']['datatype'], '100')
        self.assertEqual(options['cheese']['shape'], '0')
        self.assertEqual(options['cheese']['cheese_0_x'], '25um')
        self.assertEqual(options['cheese']['cheese_0_y'], '25um')
        self.assertEqual(options['cheese']['cheese_1_radius'], '100um')
        self.assertEqual(options['cheese']['delta_x'], '100um')
        self.assertEqual(options['cheese']['delta_y'], '100um')
        self.assertEqual(options['cheese']['edge_nocheese'], '200um')

        self.assertEqual(options['no_cheese']['datatype'], '99')
        self.assertEqual(options['no_cheese']['buffer'], '25um')
        self.assertEqual(options['no_cheese']['cap_style'], '2')
        self.assertEqual(options['no_cheese']['join_style'], '2')

        self.assertEqual(len(options['cheese']['view_in_file']), 1)
        self.assertEqual(len(options['cheese']['view_in_file']['main']), 1)
        self.assertEqual(options['cheese']['view_in_file']['main'][1], True)

        self.assertEqual(len(options['no_cheese']['view_in_file']), 1)
        self.assertEqual(len(options['no_cheese']['view_in_file']['main']), 1)
        self.assertEqual(options['no_cheese']['view_in_file']['main'][1], True)

    def test_renderer_qgmsh_renderer_options(self):
        """Test that default_options in QGmshRenderer were not accidentally
        changed."""
        design = designs.DesignPlanar()
        renderer = QGmshRenderer(design)
        options = renderer.default_options

        self.assertEqual(len(options), 4)
        self.assertEqual(len(options["mesh"]), 8)
        self.assertEqual(len(options["mesh"]["mesh_size_fields"]), 4)
        self.assertEqual(len(options["colors"]), 3)
        self.assertEqual(options["x_buffer_width_mm"], 0.2)
        self.assertEqual(options["y_buffer_width_mm"], 0.2)
        self.assertEqual(options["mesh"]["max_size"], "70um")
        self.assertEqual(options["mesh"]["min_size"], "5um")
        self.assertEqual(options["mesh"]["max_size_jj"], "5um")
        self.assertEqual(options["mesh"]["smoothing"], 10)
        self.assertEqual(options["mesh"]["nodes_per_2pi_curve"], 90)
        self.assertEqual(options["mesh"]["algorithm_3d"], 10)
        self.assertEqual(options["mesh"]["num_threads"], 8)
        self.assertEqual(
            options["mesh"]["mesh_size_fields"]["min_distance_from_edges"],
            "10um")
        self.assertEqual(
            options["mesh"]["mesh_size_fields"]["max_distance_from_edges"],
            "130um")
        self.assertEqual(options["mesh"]["mesh_size_fields"]["distance_delta"],
                         "30um")
        self.assertEqual(options["mesh"]["mesh_size_fields"]["gradient_delta"],
                         "3um")
        self.assertEqual(options["colors"]["metal"], (84, 140, 168, 255))
        self.assertEqual(options["colors"]["jj"], (84, 140, 168, 150))
        self.assertEqual(options["colors"]["dielectric"], (180, 180, 180, 255))

    def test_renderer_qelmer_renderer_options(self):
        """Test that default_options in QElmerRenderer were not accidentally
        changed."""
        design = designs.MultiPlanar()
        renderer = QElmerRenderer(design)
        options = renderer.default_options

        self.assertEqual(len(options), 6)
        self.assertEqual(options["simulation_type"], "steady_3D")
        self.assertEqual(options["simulation_dir"], "./simdata")
        self.assertEqual(options["mesh_file"], "out.msh")
        self.assertEqual(options["simulation_input_file"], "case.sif")
        self.assertEqual(options["postprocessing_file"], "case.msh")
        self.assertEqual(options["output_file"], "case.result")

    def test_renderer_ansys_renderer_get_clean_name(self):
        """Test get_clean_name in ansys_renderer.py"""
        self.assertEqual(ansys_renderer.get_clean_name('name12'), 'name12')
        self.assertEqual(ansys_renderer.get_clean_name('12name12'), 'name12')
        self.assertEqual(ansys_renderer.get_clean_name('name!'), 'name')

    def test_renderer_ansys_renderer_name_delim(self):
        """Test NAME_DELIM in QAnsysRenderer."""
        design = designs.DesignPlanar()
        renderer = QAnsysRenderer(design, initiate=False)
        self.assertEqual(renderer.NAME_DELIM, '_')

    def test_renderer_ansys_renderer_name(self):
        """Test name in QAnsysRenderer."""
        design = designs.DesignPlanar()
        renderer = QAnsysRenderer(design, initiate=False)
        self.assertEqual(renderer.name, 'ansys')

    def test_renderer_renderer_base_name(self):
        """Test name in QRenderer."""
        renderer = QRenderer
        self.assertEqual(renderer.name, 'base')

    def test_renderer_renderer_gui_base_name(self):
        """Test name in QRenderer."""
        renderer = QRendererGui
        self.assertEqual(renderer.name, 'guibase')

    def test_renderer_qgmsh_renderer_name(self):
        """Test name in QGmshRenderer."""
        design = designs.DesignPlanar()
        renderer = QGmshRenderer(design, initiate=False)
        self.assertEqual(renderer.name, 'gmsh')

    def test_renderer_qelmer_renderer_name(self):
        """Test name in QElmerRenderer."""
        design = designs.MultiPlanar()
        renderer = QElmerRenderer(design, initiate=False)
        self.assertEqual(renderer.name, 'elmer')

    def test_renderer_gdsrenderer_inclusive_bound(self):
        """Test functionality of inclusive_bound in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        my_list = []
        my_list.append([1, 1, 2, 2])
        my_list.append([3, 3, 5, 5])
        my_list.append([2.2, 2.3, 4.4, 4.9])
        self.assertEqual(renderer._inclusive_bound(my_list), (1, 1, 5, 5))

    def test_renderer_scale_max_bounds(self):
        """Test functionality of scale_max_bounds in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        actual = renderer._scale_max_bounds('main', [(1, 1, 3, 3)])
        self.assertEqual(len(actual), 2)
        self.assertEqual(actual[0], (0.8, 0.8, 3.2, 3.2))
        self.assertEqual(actual[1], (1, 1, 3, 3))

    def test_renderer_get_chip_names(self):
        """Test functionality of get_chip_names in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        qgt = QGeometryTables(design)
        qgt.clear_all_tables()

        transmon_pocket_1 = TransmonPocket(design, 'my_id')
        transmon_pocket_1.make()
        transmon_pocket_1.get_template_options(design)

        a_linestring = draw.LineString([[0, 0], [0, 1]])
        a_poly = draw.rectangle(2, 2, 0, 0)
        qgt.add_qgeometry('path',
                          'my_id', {'n_sprial': a_linestring},
                          width=4000)
        qgt.add_qgeometry('poly',
                          'my_id', {'n_spira_etch': a_poly},
                          subtract=True)

        result = renderer._get_chip_names()
        self.assertEqual(result, {'main': {}})

    def test_renderer_setup_renderers(self):
        """Test setup_renderers in setup_defauts.py."""
        actual = setup_default.setup_renderers()
        self.assertEqual(actual, {})

    def test_renderer_renderer_base_element_table_data(self):
        """Test element_table_data in QRenderer."""
        renderer = QRenderer
        etd = renderer.element_table_data

        self.assertEqual(len(etd), 0)

    def test_renderer_ansys_renderer_element_table_data(self):
        """Test element_table_data in QAnsysRenderer."""
        design = designs.DesignPlanar()
        renderer = QAnsysRenderer(design, initiate=False)
        etd = renderer.element_table_data

        self.assertEqual(len(etd), 2)
        self.assertEqual(len(etd['path']), 1)
        self.assertEqual(len(etd['junction']), 4)

        self.assertEqual(etd['path']['wire_bonds'], False)
        self.assertEqual(etd['junction']['inductance'], '10nH')
        self.assertEqual(etd['junction']['capacitance'], 0)
        self.assertEqual(etd['junction']['resistance'], 0)
        self.assertEqual(etd['junction']['mesh_kw_jj'], 7e-06)

    def test_renderer_gdsrenderer_high_level(self):
        """Test that high level defaults were not accidentally changed in
        gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        self.assertEqual(renderer.name, 'gds')
        element_table_data = renderer.element_table_data
        self.assertEqual(len(element_table_data), 1)
        self.assertEqual(len(element_table_data['junction']), 1)
        self.assertEqual(element_table_data['junction']['cell_name'],
                         'my_other_junction')

    def test_renderer_gdsrenderer_update_units(self):
        """Test update_units in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        renderer.options['gds_unit'] = 12345
        self.assertEqual(renderer.options['gds_unit'], 12345)

        renderer._update_units()
        self.assertEqual(renderer.options['gds_unit'], 0.001)

    def test_renderer_gdsrenderer_midpoint_xy(self):
        """Test midpoint_xy in gds_renderer.py."""
        actual = QGDSRenderer._midpoint_xy(10, 15, 20, 30)
        self.assertEqual(len(actual), 2)
        self.assertEqual(actual[0], 15.0)
        self.assertEqual(actual[1], 22.5)

    # pylint: disable-msg=unused-variable
    def test_renderer_gdsrenderer_check_qcomps(self):
        """Test check_qcomps in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        actual = []
        actual.append(renderer._check_qcomps([]))
        actual.append(renderer._check_qcomps(['Q1', 'Q2', 'Q3']))
        actual.append(renderer._check_qcomps(['Q1', 'Q2', 'Q3', 'Q1']))

        expected = []
        expected.append(([], 0))
        expected.append((['Q2', 'Q3', 'Q1'], 1))
        expected.append((['Q2', 'Q3', 'Q1'], 1))

        for x in range(3):
            my_length = len(actual[x][0])
            self.assertEqual(my_length, len(expected[x][0]))
            self.assertEqual(actual[x][1], expected[x][1])

            for y, _ in enumerate(expected[x][0]):
                self.assertTrue(_ in actual[x][0])

    def test_renderer_mpl_interaction_disconnect(self):
        """Test disconnect in MplInteraction in mpl_interaction.py."""
        mpl = MplInteraction(_plt)
        mpl.disconnect()
        self.assertEqual(mpl.figure, None)

    def test_renderer_gds_check_cheese(self):
        """Test check_cheese in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        self.assertEqual(renderer._check_cheese('main', 0), 4)
        self.assertEqual(renderer._check_cheese('main', 1), 1)
        self.assertEqual(renderer._check_cheese('fake', 0), 3)

    def test_renderer_gds_check_no_cheese(self):
        """Test check_no_cheese in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        self.assertEqual(renderer._check_no_cheese('main', 0), 4)
        self.assertEqual(renderer._check_no_cheese('main', 1), 1)
        self.assertEqual(renderer._check_no_cheese('fake', 0), 3)

    def test_renderer_gds_check_either_cheese(self):
        """Test check_either_cheese in gds_renderer.py."""
        design = designs.DesignPlanar()
        renderer = QGDSRenderer(design)

        self.assertEqual(renderer._check_either_cheese('main', 0), 6)
        self.assertEqual(renderer._check_either_cheese('main', 1), 1)
        self.assertEqual(renderer._check_either_cheese('fake', 0), 5)

    def test_successful_get_convergence(self):
        """Test get_convergence returns the correct Boolean when converged"""

        convergence_txt = '''Setup : Setup
        \nProblem Type : CG\n
        \n==================\n
        Number of Passes\n
        Completed : 15\n
        Maximum   : 20\n
        Minimum   : 2\n
        ==================\n
        Criterion : Delta %\n
        Target    : 0.1\n
        Current   : 0.042998\n
        ==================\n
        Pass|# Triangle| Delta %|\n
        1|       108|     N/A|\n
        2|       326|   1.739|\n
        3|       966|  1.3044|\n
        4|      2358|  1.2871|\n
        5|      3040|  1.0324|\n
        6|      4038| 0.53456|\n
        7|      5470| 0.50188|\n
        8|      7440| 0.28342|\n
        9|     10122| 0.33775|\n
        10|     13768| 0.10967|\n
        11|     18498| 0.11437|\n
        12|     25044| 0.13151|\n
        13|     33562| 0.13472|\n
        14|     44028| 0.07696|\n
        15|     57390|0.042998|\n\n'''
        design = designs.DesignPlanar()
        renderer = QQ3DRenderer(design, initiate=False)
        renderer._pinfo = MagicMock()
        renderer._pinfo.setup = MagicMock()
        renderer._pinfo.setup.get_convergence = MagicMock()
        renderer._pinfo.setup.get_convergence.return_value = (None,
                                                              convergence_txt)
        self.assertTrue(renderer.get_convergence())

    def test_unsuccessful_get_convergence(self):
        """Test get_convergence returns the correct Boolean when not converged"""

        convergence_txt = ''''Setup : Setup\n
            Problem Type : CG\n\n
            ==================\n
            Number of Passes\n
            Completed : 2\n
            Maximum   : 2\n
            Minimum   : 2\n
            ==================\n
            Criterion : Delta %\n
            Target    : 0.1\n
            Current   : 1.739\n
            ==================\n
            Pass|# Triangle|Delta %|\n
            1|       108|    N/A|\n
            2|       326|  1.739|\n\n'''
        design = designs.DesignPlanar()
        renderer = QQ3DRenderer(design, initiate=False)
        renderer._pinfo = MagicMock()
        renderer._pinfo.setup = MagicMock()
        renderer._pinfo.setup.get_convergence = MagicMock()
        renderer._pinfo.setup.get_convergence.return_value = (None,
                                                              convergence_txt)
        self.assertFalse(renderer.get_convergence())


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6 import QtCore, QtGui
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QToolBar
import time


class QToolBarExpanding(QToolBar):
    """`QToolBarExpanding` class extends the `QToolBar` class.

    Example:
        ```toolbar = gui.ui.toolBarView```

    Args:
        QToolbar (QToolbar): QToolbar
    """

    def expand_me(self):
        """Expand the toolbar."""
        if self.orientation() == Qt.Vertical:
            tool_style = Qt.ToolButtonTextBesideIcon
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:  # Qt.Horizontal
            tool_style = Qt.ToolButtonTextUnderIcon
            align = Qt.AlignHCenter | Qt.AlignTop

        # show icons and text
        self.setToolButtonStyle(tool_style)

        # align icons and text
        layout = self.layout()
        layout.setAlignment(align)
        layout.setSpacing(layout.spacing())
        for i in range(layout.count()):
            tool = layout.itemAt(i)
            tool.setAlignment(align)
            # https://doc.qt.io/qt-5/qlayoutitem.html#setAlignment

    def contract_me(self):
        """Contract the toolbar."""
        self.setToolButtonStyle(Qt.ToolButtonIconOnly)

    def enterEvent(self, evt: QtCore.QEvent) -> None:
        """enterEvent() is called when the mouse enters the widget's screen
        space. (This excludes screen space owned by any of the widget's
        children.)

        Args:
            evt (QtCore.QEvent): QtCore event
        """
        # should ideally have a timeout thread
        # print('-> Enter')
        self.expand_me()

        super().enterEvent(evt)

    def leaveEvent(self, evt: QtCore.QEvent) -> None:
        """leaveEvent() is called when the mouse leaves the widget's screen
        space. If the mouse enters a child widget it will not cause a
        leaveEvent().

        Args:
            evt (QtCore.QEvent): QtCore event
        """
        # print('<- EXIT')
        # this adds a 0.25sec delay before contracting the widget if mouse accidentally leaves
        time.sleep(0.25)
        self.contract_me()
        super().leaveEvent(evt)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Placeholder to allow for future modifications of the attribute dictionary
class.

We use the `addict` package `Dict` class.
"""

# pylint: disable=unused-import

from addict import Dict



"""This code calculates the photon loss (kappa) due to the capacitive coupling
between CPWs and input/output transmission lines in a quantum circuit.

Two cases are treated: In the first case, three arguments are passed to the function kappa_in
and the resonant frequency of the CPW is input as a float. In the second case, six arguments
are passed to kappa_in and the frequency of the CPW is calculated assuming an ideal CPW.

Key References:

D. Schuster, Ph.D. Thesis, Yale University (2007)
https://rsl.yale.edu/sites/default/files/files/RSL_Theses/SchusterThesis.pdf

T. McConkey, Ph.D. Thesis, University of Waterloo (2018)
https://uwspace.uwaterloo.ca/bitstream/handle/10012/13464/McConkey_Thomas.pdf?sequence=3&isAllowed=y

Mohebbi and Majedi, Superconducting Science and Technology 22, 125028 (2009)
https://iopscience.iop.org/article/10.1088/0953-2048/22/12/125028/meta

P. Krantz, et al. Physical Review Applied 6, 021318 (2019)
https://aip.scitation.org/doi/10.1063/1.5089550
"""

import numpy as np
from math import *
from scipy.special import ellipk

__all__ = ['kappa_in']


def kappa_in(*argv):
    """A simple calculator for the kappa value of a readout resonator.

    Args:
        freq (float): The frequency of interest, in Hz
        C_in (float): Effective capacitance between CPW and environment (from Q3D), in Farads
        freq_res (float): Lowest resonant frequency of a CPW (from HFSS), in Hz
        length (float): Length of the CPW readout resonator, in meters
        res_width (float): Width of the resonator trace (center) line, in meters
        res_gap (float): Width of resonator gap (dielectric space), in meters
        eta (float): 2.0 for half-wavelength resonator; 4.0 for quarter-wavelength resonator

    Returns:
        float: Kappa value
    """

    # Effective impedance of the CPW transmission line, in Ohms
    Z_tran = 50.0

    # Effective impedance of the readout resonator, in Ohms
    Z_res = 50.0

    # If three arguments are passed to kappa_in, then the lowest resonator frequency is assumed to be an input
    if len(argv) == 3:
        for i in argv:
            freq = argv[0]
            C_in = argv[1]
            freq_res = argv[2]

        # Calculation of kappa
        kappa = (2 / pi) * (freq**2.0) * (C_in**2.0) * (Z_tran**
                                                        2.0) * (freq_res)

        return kappa

    # If six arguments are passed to kappa_in, the lowest resonator frequency of the resonator is calculated in the ideal case
    elif len(argv) == 6:
        for i in argv:
            freq = argv[0]
            C_in = argv[1]
            length = argv[2]
            res_width = argv[3]
            res_gap = argv[4]
            eta = argv[5]

        # Arguments for elliptic integrals
        k0 = (res_width) / (res_width + 2.0 * res_gap)
        k01 = (1.0 - k0**2.0)**(0.5)

        # Calculation of the first resonant frequency of an ideal resonator
        freq_res = (Z_res) * (ellipk(
            k0**2.0)) / (15.0 * eta * length * ellipk(k01**2.0))

        # Calculation of kappa
        kappa = (2 / pi) * (freq**2.0) * (C_in**2.0) * (Z_tran**
                                                        2.0) * (freq_res)
        return kappa

    # Only three or six arguments accepted by kappa_in, otherwise the calculation in invalid.
    else:
        kappa = "Invalid number of arguments passed for kappa_in"



import sequencing as seq
import numpy as np
import scqubits as scq
from typing import List

from qiskit_metal.analyses.quantization.lom_core_analysis import CompositeSystem, Subsystem
from qiskit_metal.analyses.quantization.lom_extensions import LOM_SUBSYSTEM_TO_SEQ_MODE, to_external_system


def lom_composite_sys_to_seq_sys(lom_composite: CompositeSystem,
                                 hilbertspace: scq.HilbertSpace,
                                 levels: List[int] = None,
                                 system_name='system') -> seq.System:
    ## TODO: params for setting detuning and levels
    """Convert LOM composite system to Sequencing system. A 'subsystem' in
    LOM composite system corresponds to a 'mode' in Sequencing system

    Args:
        lom_composite (CompositeSystem): Metal LOM composite system
        hilbertspace (scq.HilbertSpace): The hilbertspace of the passed LOM composite system
        levels (List[int], optional): Number of energy levels to keep for each mode in
            Sequencing system. If None, use the LOM subsystems' respective dimensions.
            Defaults to None.
        system_name (str, optional): [description]. Defaults to 'system'.

    Returns:
        seq.System: the converted Sequencing system
    """

    modes = []

    if levels is None:
        levels = hilbertspace.subsystem_dims

    h_results = lom_composite.hamiltonian_results(hilbertspace,
                                                  print_info=False)
    chi_mat = h_results['chi_in_MHz']
    chi_mat_ghz = chi_mat * 1e-3
    for ii, sub1 in enumerate(lom_composite.subsystems):
        seq_cls = to_external_system(sub1, LOM_SUBSYSTEM_TO_SEQ_MODE)
        sub_name = sub1.name
        self_kerr = chi_mat_ghz[ii, ii]
        mode = seq_cls(sub_name, levels=levels[ii], kerr=self_kerr)
        modes.append(mode)

    system = seq.System(system_name, modes=modes)
    for ii in range(len(modes)):
        mode1 = modes[ii]
        for jj in range(ii + 1, len(modes)):
            mode2 = modes[jj]
            cross_kerr = chi_mat_ghz[ii, jj]
            system.set_cross_kerr(mode1, mode2, chi=cross_kerr)

    return system



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=invalid-name
# pylint: disable-msg=bad-continuation
"""Custom assertion for unit tests."""

from itertools import zip_longest
from typing import Iterable, Optional


class AssertionsMixin:
    """Custom assertions for unit tests."""

    def assertAlmostEqualRel(
        self,
        expected: float,
        tested: float,
        *,
        msg: Optional[str] = None,
        rel_tol: float = 1e-7,
        abs_tol: float = 0.0,
    ):
        """Assert tested almost equals expected.

        Args:
            expected (float): Expected value to compare to.
            tested (float): Value to compare to expected.
            msg (Optional[str]): Custom message to show on failure.  Defaults to None.
            rel_tol (float): Relative tolerance to accept.  Defaults to 1e-7.
            abs_tol (float): Absolute tolerance to accept.  Defaults to 0.0.

        Return:
            Exception: Failure exception, or nothing
        """
        if expected == tested:
            return

        rel_diff = abs((expected - tested) / expected)
        abs_diff = abs(expected - tested)

        if rel_diff <= rel_tol or abs_diff <= abs_tol:
            return

        if rel_diff > rel_tol:
            standardMsg = (
                f"Relative difference {rel_diff} exceeds tolerance {rel_tol} "
                f"for expected={expected}, tested={tested}")
            msg = self._formatMessage(msg, standardMsg)
            raise self.failureException(msg)

        if abs_diff > abs_tol:
            standardMsg = (
                f"Absolute difference {abs_diff} exceeds tolerance {abs_tol} "
                f"for expected={expected}, tested={tested}")
            msg = self._formatMessage(msg, standardMsg)
            raise self.failureException(msg)

    def assertIterableAlmostEqual(
        self,
        expected: Iterable[float],
        tested: Iterable[float],
        msg: Optional[str] = None,
        rel_tol: float = 1e-7,
        abs_tol: float = 0.0,
    ):
        """Assert all entries in two iterables are almost equal.

        Args:
            expected (Iterable[float]): First iterable to compare.
            tested (Iterable[float]): Second iterable to compare.
            msg (Optional[str]): Custom message to show for failure.  Defaults to None.
            rel_tol (float): Relative tolerance to accept.  Defaults to 1e-7.
            abs_tol (float): Absolute tolerance to accept.  Defaults to 0.0.

        Return:
            Exception: Failure exception, or nothing
        """
        sentinel = object()
        for index, (item1, item2) in enumerate(
                zip_longest(expected, tested, fillvalue=sentinel)):
            if sentinel in (item1, item2):
                standardMsg = (
                    f"Iterables have unequal lengths. One has {index} and the "
                    "other has more.")
                msg = self._formatMessage(msg, standardMsg)
                raise self.failureException(msg)

            try:
                self.assertAlmostEqualRel(item1,
                                          item2,
                                          rel_tol=rel_tol,
                                          abs_tol=abs_tol)
            except AssertionError as err:
                standardMsg = f"Iterable elements {index} not almost equal: {str(err)}"
                msg = self._formatMessage(msg, standardMsg)
                raise self.failureException(msg) from None



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
""""""

import matplotlib.pyplot as plt
from PySide6.QtCore import QTimer

from ...._gui.utility._handle_qt_messages import slot_catch_error


class AnimatedText():
    """Class that animates text."""

    def __init__(self,
                 ax: plt.Axes,
                 text: str,
                 canvas,
                 kw: dict = None,
                 anim_start=0.9,
                 anim_dt_ms=25,
                 anim_delta=-0.0005,
                 anim_stop=0,
                 anim_accel=-0.0005,
                 start=True,
                 loc=[0.5, 0.5]):
        """
        Args:
            ax (plt.Axes): The axis.
            text (str): Text to animate.
            canvas (canvas): The canvas.
            kw (dict): The parameters.  Defaults to None.
            anim_start (float): Animation start.  Defaults to 0.9.
            anim_dt_ms (int): Animation dt in miliseconds.  Defaults to 25.
            anim_delta (float): Animation delta.  Defaults to -0.0005.
            anim_stop (int): Animation stop.  Defaults to 0.
            anim_accel (float): Animation acceleration.  Defaults to -0.0005.
            start (bool): Whether or not to start.  Defaults to True.
            loc (list): Location.  Defaults to [0.5, 0.5].
        """

        self.canvas = canvas
        self.ax = ax
        self.anim_value = anim_start
        self.anim_delta = anim_delta
        self.anim_dt_ms = anim_dt_ms
        self.anim_stop = anim_stop
        self.anim_accel = anim_accel
        self.anim_veloc = 0

        # MPL text
        # Create the text
        kw = {
            **dict(fontsize=35,
                   fontweight='bold',
                   va='center',
                   ha='center',
                   alpha=anim_start,
                   transform=ax.transAxes,
                   color='#00356B',
                   zorder=200),
            **(kw if kw else {})
        }

        self.text = ax.text(*loc, text, **kw)

        # Timer
        self.timer = QTimer()
        self.timer.timeout.connect(self.timer_tick)

        if start:
            self.start()

    def start(self):
        """Start the timer."""
        self.timer.start(self.anim_dt_ms)

    def stop(self):
        """Stop the timer."""
        self.timer.start()

    @slot_catch_error()
    def timer_tick(self):
        """Tick the timer."""
        # Update anim position value
        self.anim_veloc += self.anim_accel  # acceleration on vellcoity update
        self.anim_value += self.anim_delta + self.anim_veloc  # update position
        # print(f'tick {self.anim_value}')

        if self.anim_value >= self.anim_stop:  # one sided
            self.text.set_alpha(self.anim_value)
            if self.text not in self.ax.texts:
                # if the axis is cleared and redrawn
                self.ax.texts.append(self.text)
            self.canvas.refresh()  # refresht the canvas

        else:
            self.timer.stop()
            self.timer.deleteLater()

            # Remove artist
            self.text.figure = self.ax.figure
            if self.text in self.ax.texts:
                self.ax.texts.remove(self.text)

            if self.text.axes:
                try:
                    # remove text from axis
                    self.text.remove()
                except ValueError as e:
                    # could raise ValueError: list.remove(x): x not in list
                    pass

            self.canvas.refresh()  # refresht the canvas



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=broad-except
# pylint: disable-msg=import-outside-toplevel
# pylint: disable-msg=import-error
"""Qiskit Metal unit tests analyses functionality."""

import unittest
import numpy as np

from geopandas import GeoDataFrame

from qiskit_metal import designs
from qiskit_metal import draw

from qiskit_metal.qgeometries import qgeometries_handler
from qiskit_metal.qgeometries.qgeometries_handler import QGeometryTables
from qiskit_metal.qlibrary.qubits.transmon_pocket import TransmonPocket


class TestElements(unittest.TestCase):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_qgeometry_instantiate_q_geometry_tables(self):
        """Test instantiation of QGeometryTables."""
        try:
            design = designs.DesignPlanar()
            QGeometryTables(design)
        except Exception:
            self.fail("QGeometryTables failed")

    def test_qgeometry_is_qgeometry_table(self):
        """Test is_qgeometry_table in qgeometries_handler.py."""
        self.assertEqual(
            qgeometries_handler.is_qgeometry_table(QGeometryTables), True)
        self.assertEqual(
            qgeometries_handler.is_qgeometry_table("i_am_a_string"), False)

    def test_qgeometry_element_columns(self):
        """Test that ELEMENT_COLUMNS was not accidentally changed in
        element_handler.py."""
        import importlib
        importlib.reload(qgeometries_handler)
        e_c = qgeometries_handler.ELEMENT_COLUMNS

        self.assertEqual(len(e_c), 4)

        self.assertEqual(len(e_c['base']), 8)
        self.assertEqual(e_c['base']['component'], str)
        self.assertEqual(e_c['base']['name'], str)
        self.assertEqual(e_c['base']['geometry'], object)
        self.assertEqual(e_c['base']['layer'], int)
        self.assertEqual(e_c['base']['subtract'], bool)
        self.assertEqual(e_c['base']['helper'], bool)
        self.assertEqual(e_c['base']['chip'], str)
        self.assertEqual(e_c['base']['__renderers__'], dict())
        self.assertEqual(len(e_c['base']['__renderers__']), 0)

        self.assertEqual(len(e_c['path']), 3)
        self.assertEqual(e_c['path']['width'], float)
        self.assertEqual(e_c['path']['fillet'], object)
        self.assertEqual(e_c['path']['__renderers__'], dict())
        self.assertEqual(len(e_c['path']['__renderers__']), 0)

        self.assertEqual(len(e_c['poly']), 2)
        self.assertEqual(e_c['poly']['fillet'], object)
        self.assertEqual(e_c['poly']['__renderers__'], dict())
        self.assertEqual(len(e_c['poly']['__renderers__']), 0)

        self.assertEqual(len(e_c['junction']), 2)
        self.assertEqual(e_c['junction']['width'], float)
        self.assertEqual(e_c['junction']['__renderers__'], dict())

    # pylint: disable-msg=unused-variable
    def test_qgeometry_true_bools(self):
        """Test that TRUE_BOOLS was not accidentally changed in
        element_handler.py."""
        expected = [True, 'True', 'true', 'Yes', 'yes', '1', 1]
        actual = qgeometries_handler.TRUE_BOOLS

        self.assertEqual(len(actual), len(expected))

        for x, _ in enumerate(expected):
            self.assertTrue(_ in actual)

    def test_qgeometry_q_element_constants(self):
        """Test that constants in QGeometryTables class in element_handler.py
        were not accidentally changed."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)

        self.assertEqual(qgt.__i_am_qgeometry_table__, True)
        self.assertEqual(qgt.name_delimiter, '_')

    def test_qgeometry_q_element_add_renderer_extension(self):
        """Test add_renderer_extension in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.add_renderer_extension(
            'new_name',
            dict(base=dict(color=str, klayer=int),
                 path=dict(thickness=float),
                 poly=dict(material=str)))

        e_c = qgt.ELEMENT_COLUMNS

        self.assertTrue('color' in e_c['base']['__renderers__']['new_name'])
        self.assertTrue('klayer' in e_c['base']['__renderers__']['new_name'])
        self.assertTrue('thickness' in e_c['path']['__renderers__']['new_name'])
        self.assertTrue('material' in e_c['poly']['__renderers__']['new_name'])

        self.assertEqual(e_c['base']['__renderers__']['new_name']['color'], str)
        self.assertEqual(e_c['base']['__renderers__']['new_name']['klayer'],
                         int)
        self.assertEqual(e_c['path']['__renderers__']['new_name']['thickness'],
                         float)
        self.assertEqual(e_c['poly']['__renderers__']['new_name']['material'],
                         str)

    def test_qgeometry_q_element_get_element_types(self):
        """Test get_element_types in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)

        expected = ['path', 'poly', 'junction']
        actual = qgt.get_element_types()

        self.assertEqual(len(expected), len(actual))

        for x, _ in enumerate(expected):
            self.assertTrue(_ in actual)

    def test_qgeometry_q_element_create_tables(self):
        """Test create_tables in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.create_tables()

        actual = qgt.tables

        self.assertEqual(len(actual), 3)
        self.assertTrue('path' in actual)
        self.assertTrue('poly' in actual)
        self.assertTrue('junction' in actual)

        self.assertEqual(actual['path'].dtypes['component'], object)
        self.assertEqual(actual['path'].dtypes['name'], object)
        self.assertEqual(actual['path'].dtypes['subtract'], bool)
        self.assertEqual(actual['path'].dtypes['helper'], bool)
        self.assertEqual(actual['path'].dtypes['chip'], object)
        self.assertEqual(actual['path'].dtypes['fillet'], object)

        self.assertEqual(actual['poly'].dtypes['component'], object)
        self.assertEqual(actual['poly'].dtypes['name'], object)
        self.assertEqual(actual['poly'].dtypes['subtract'], bool)
        self.assertEqual(actual['poly'].dtypes['helper'], bool)
        self.assertEqual(actual['poly'].dtypes['chip'], object)
        self.assertEqual(actual['poly'].dtypes['fillet'], object)

        self.assertEqual(actual['junction'].dtypes['component'], object)
        self.assertEqual(actual['junction'].dtypes['name'], object)
        self.assertEqual(actual['junction'].dtypes['subtract'], bool)
        self.assertEqual(actual['junction'].dtypes['helper'], bool)
        self.assertEqual(actual['junction'].dtypes['chip'], object)
        self.assertEqual(actual['junction'].dtypes['width'], float)

    def test_qgeometry_q_element_get_rname(self):
        """Test get_rname in QGeometryTables class in element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)

        actual = qgt.get_rname('some_name', 'aloha')

        self.assertEqual(actual, 'some_name_aloha')

    def test_qgeometry_q_element_add_qgeometry(self):
        """Test add_qgeometry in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.clear_all_tables()

        a_poly = draw.rectangle(2, 2, 0, 0)
        qgt.add_qgeometry('poly', 'my_id', dict(cl_metal=a_poly))
        table = qgt.tables

        self.assertTrue('poly' in table)
        self.assertEqual(table['poly']['component'][0], 'my_id')
        self.assertEqual(table['poly']['name'][0], 'cl_metal')
        self.assertEqual(table['poly']['geometry'][0], a_poly)
        self.assertEqual(table['poly']['layer'][0], 1)
        self.assertEqual(table['poly']['subtract'][0], False)
        self.assertEqual(table['poly']['helper'][0], False)
        self.assertEqual(table['poly']['chip'][0], 'main')
        self.assertEqual(str(table['poly']['fillet'][0]), str(np.nan))

    def test_qgeometry_q_element_clear_all_tables(self):
        """Test clear_all_tables in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.clear_all_tables()

        # add something to the tables to check for after clear
        a_poly = draw.rectangle(2, 2, 0, 0)
        qgt.add_qgeometry('poly', 'my_id', dict(cl_metal=a_poly))

        qgt.clear_all_tables()
        self.assertEqual(len(qgt.tables['poly'].geometry), 0)

    def test_qgeometry_q_element_delete_component_id(self):
        """Test delete_component_id in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.clear_all_tables()

        transmon_pocket = TransmonPocket(design, 'my_id')
        transmon_pocket.make()
        transmon_pocket.get_template_options(design)
        a_linestring = draw.LineString([[0, 0], [0, 1]])
        a_poly = draw.rectangle(2, 2, 0, 0)
        qgt.add_qgeometry('path',
                          'my_id', {'n_sprial': a_linestring},
                          width=4000)
        qgt.add_qgeometry('poly',
                          'my_id', {'n_spira_etch': a_poly},
                          subtract=True)

        self.assertEqual(len(qgt.tables['path']), 1)
        self.assertEqual(len(qgt.tables['poly']), 1)

        qgt.delete_component_id('my_id')

        self.assertEqual(len(qgt.tables['path']), 0)
        self.assertEqual(len(qgt.tables['poly']), 0)

    def test_qgeometry_get_all_unique_layers(self):
        """Test get_all_unique_layers functionality in elment_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.clear_all_tables()

        transmon_pocket = TransmonPocket(design, 'my_id')
        transmon_pocket.make()
        transmon_pocket.get_template_options(design)
        a_linestring = draw.LineString([[0, 0], [0, 1]])
        a_poly = draw.rectangle(2, 2, 0, 0)
        qgt.add_qgeometry('path',
                          'my_id', {'n_sprial': a_linestring},
                          width=4000)
        qgt.add_qgeometry('poly',
                          'my_id', {'n_spira_etch': a_poly},
                          subtract=True)

        self.assertEqual(qgt.get_all_unique_layers('main'), [1])
        self.assertEqual(qgt.get_all_unique_layers('fake'), [])

    def test_qgeometry_q_element_get_component_bounds(self):
        """Test get_component_bounds in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.clear_all_tables()
        TransmonPocket(design, 'my_id')

        four_zeros = qgt.get_component_bounds('my_id')
        self.assertEqual(len(four_zeros), 4)
        for i in range(4):
            self.assertEqual(four_zeros[i], 0)

    def test_qgeometry_q_element_check_element_type(self):
        """Test check_element_type in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        TransmonPocket(design, 'my_id')

        self.assertTrue(qgt.check_element_type('path', log_issue=False))
        self.assertFalse(qgt.check_element_type('not-there', log_issue=False))

    def test_qgeometry_q_element_get_component(self):
        """Test get_component in QGeometryTables class in
        element_handler.py."""
        design = designs.DesignPlanar()
        qgt = QGeometryTables(design)
        qgt.clear_all_tables()
        TransmonPocket(design, 'Q1')

        rect = draw.rectangle(500, 300, 0, 0)
        geom = {'my_polygon': rect}
        qgt.add_qgeometry('poly', 'Q1', geom)

        # success results
        actual = qgt.get_component('Q1')
        self.assertEqual(len(actual), 3)
        self.assertTrue(isinstance(actual['path'], GeoDataFrame))
        self.assertTrue(isinstance(actual['poly'], GeoDataFrame))
        self.assertTrue(isinstance(actual['junction'], GeoDataFrame))

        # failure results
        actual = qgt.get_component('not-real')
        self.assertEqual(len(actual), 3)
        self.assertEqual(actual['path'], None)
        self.assertEqual(actual['poly'], None)
        self.assertEqual(actual['junction'], None)


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

#pylint: disable-msg=import-error
#pylint: disable-msg=consider-using-enumerate
"""
This code calculates the energy eigenvalues of the transmon qubit cooper pair box
as a function of the offset charge and plots the corresponding results.

Key References:
    - J. Koch et al. "Charge-insensitive qubit design derived from the Cooper pair box"
      Phys. Rev. A. 76, 042319 (2007).

"""

from scipy.special import mathieu_a
import numpy as np
import matplotlib.pyplot as plt

__all__ = ['kidx_raw', 'kidx', 'transmon_eigenvalue', 'plot_eigenvalues']


def kidx_raw(m, my_ng):
    """
    This function calculates the integers which correct sort the eigenstates and eigenenergies
    of the solution to Mathieu's equation.

        Args:
            m (int): The energy level of the qubit (m=0,1,2,3,etc.)
            my_ng (float): the offset charge of the Josephjunction island (in units of 2e)

        Returns:
            float: the calculated index
    """
    if my_ng == 0:
        return m + 1.0 - ((m + 1.0) % 2.0)

    return m + 1.0 - ((m + 1.0) % 2.0) + 2.0 * my_ng * (
        (-1.0)**(m - 0.5 * (np.sign(my_ng) - 1.0)))


def kidx(m, my_ng):
    """
    This function is a placeholder which performs an additional modulo operation
    with an offset of ng=0.5, but is not necessary when only integer values of
    kidxRAW(m, my_ng) are used. If non-integer values of kidxRAW are used, then
    the modulo operation below needs to be updated to include the offset.

        Args:
            m (int): The energy level of the qubit (m=0,1,2,3,etc.)
            ng (float): the offset charge of the Josephjunction island (in units of 2e)

        Returns:
            float: the calculated energy eigenvalue, after an ng=0.5 offset.
    """
    return kidx_raw(m, my_ng)


# define the Josephson to charging energy ratio E_J / E_C
RATIO = 1.0

# define the charing energy
E_C = 1.0


def transmon_eigenvalue(m, my_ng):
    """
    This function calculate the energy eigenvalue of the transmon qubit for a given
    energy level (m) and offset charge (my_ng). The input values are first used to
    calculate the index using the function defined above, and then the calculated
    index is used to calculate the energy eigenvalue using Mathieu's characteristic values.

        Args:
            m (int): The energy level of the qubit (m=0,1,2,3,etc.)
            ng (float): the offset charge of the Josephjunction island (in units of 2e)

        Returns:
            float: the calculated energy eigenvalue.
    """
    index = kidx(m, my_ng)
    return (E_C) * mathieu_a(index, -0.5 * RATIO)


# extremely coarse grid: only three points from -0.5 to 0.5. This so that only integer values
# of the index are used in the calculation of the Mathieu characteristic value.
ng = np.linspace(-0.5, 0.5, 3)

# ng is periodic extending from -2 to 2:
ng_periodic = np.linspace(-2.0, 2.0, 9)

# define energies between (-0.5, 0.5) as empty lists.
E0 = []
E1 = []
E2 = []
E3 = []

# define periodic energies between (-2.0, 2.0)
E0_periodic = [None] * 9
E1_periodic = [None] * 9
E2_periodic = [None] * 9
E3_periodic = [None] * 9

# calculate the energies for m=0,1,2 at each value of offset charge
for i in ng:
    E0.append(transmon_eigenvalue(0, i))
    E1.append(transmon_eigenvalue(1, i))
    E2.append(transmon_eigenvalue(2, i))
    E3.append(transmon_eigenvalue(3, i))

# define the periodic eigen energies based on the values between (-0.5,0.5)
for i in range(len(E0_periodic)):
    E0_periodic[0] = E0[1]
    E0_periodic[1] = E0[0]
    E1_periodic[0] = E1[1]
    E1_periodic[1] = E1[0]
    E2_periodic[0] = E2[1]
    E2_periodic[1] = E2[0]
    E3_periodic[0] = E3[1]
    E3_periodic[1] = E3[0]

    if i > 1:
        E0_periodic[i] = E0_periodic[i - 2]
        E1_periodic[i] = E1_periodic[i - 2]
        E2_periodic[i] = E2_periodic[i - 2]
        E3_periodic[i] = E3_periodic[i - 2]


def plot_eigenvalues():
    """
    This function actually creates the plot(s) of eigenvalues as a function of
    offset charge. No arguments need to be passed.

    Returns:
        A plot of the eigenvalues as a function of offset charge.
    """
    # plot the PERIODIC eigen energies between (-2.0, 2.0)
    plt.plot(ng_periodic, E0_periodic, 'k')  # m=0
    plt.plot(ng_periodic, E1_periodic, 'r')  # m=1
    plt.plot(ng_periodic, E2_periodic, 'b')  # m=2
    plt.plot(ng_periodic, E3_periodic, 'm')  # m=3
    plt.xlabel("Offset Charge [ng]")
    plt.ylabel("Energy E_m[ng]")



# -*- coding: utf-8 -*-
from re import sub
from typing import List, Tuple, Union
from copy import deepcopy

import pandas as pd


def determine_larger_box(minx: Union[None, float], miny: Union[None, float],
                         maxx: Union[None, float], maxy: Union[None, float],
                         chip_box: tuple) -> Tuple[float, float, float, float]:
    """Return box which includes the two boxes.

    Args:
        minx (Union[None, float]): Minimum of x coordinate
        miny (Union[None, float]): Minimum of y coordinate
        maxx (Union[None, float]): Maximum of x coordinate
        maxy (Union[None, float]): Maximum of y coordinate
        chip_box (tuple):  Second box in following format:
                            minx, miny, maxx, maxy

    Returns:
        Tuple[float, float, float, float]: The size:  minx, miny, maxx, maxy of
                                        box which includes both boxes.
    """

    large_minx, large_miny, large_maxx, large_maxy = None, None, None, None

    if minx and miny and maxx and maxy:
        chip_minx, chip_miny, chip_maxx, chip_maxy = chip_box
        large_minx = min(minx, chip_minx)
        large_miny = min(miny, chip_miny)
        large_maxx = max(maxx, chip_maxx)
        large_maxy = max(maxy, chip_maxy)
    else:
        # First time getting chip size, so just use chip_box
        large_minx, large_miny, large_maxx, large_maxy = chip_box

    return large_minx, large_miny, large_maxx, large_maxy


class BoundsForPathAndPolyTables():
    """Create class which can be used by multiple renderers.  In particular, this class
    assumes a LayerStack is being used within QDesign.
    """

    def __init__(self, design: 'MultiPlanar'):
        self.design = design
        self.chip_names_matched = None  # bool
        self.valid_chip_names = None  # set of valid chip names from layer_stack

    def get_bounds_of_path_and_poly_tables(
        self, box_plus_buffer: bool, qcomp_ids: List, case: int, x_buff: float,
        y_buff: float
    ) -> tuple[tuple[float, float, float, float], pd.DataFrame, pd.DataFrame,
               bool, Union[None, set]]:
        """Return bounds and concatenated geometry tables for selected components.

        Args:
            box_plus_buffer: If True, use selected components' bounding box plus
                the supplied buffer; otherwise use the chip size.
            qcomp_ids: Component ids to include (may be empty to include all).
            case: Return code from ``QRenderer.get_unique_component_ids``.
            x_buff: Buffer to add in x when ``box_plus_buffer`` is True.
            y_buff: Buffer to add in y when ``box_plus_buffer`` is True.

        Returns:
            tuple: Bounding box (minx, miny, maxx, maxy).
            pd.DataFrame: Path and poly tables concatenated for the selection.
            pd.DataFrame: Path, poly, and junction tables concatenated.
            bool: True if chip names match between design and layer stack.
            set | None: Valid chip names if they match, otherwise None.
        """
        box_for_xy_bounds = None
        self.chip_names_matched = None
        self.valid_chip_names = None

        path_dataframe = self.design.qgeometry.tables['path']
        poly_dataframe = self.design.qgeometry.tables['poly']
        junction_dataframe = self.design.qgeometry.tables['junction']

        # NOTE:get_box_for_xy_bounds populates self.chip_names_matched and self.valid_chip_names
        chip_minx, chip_miny, chip_maxx, chip_maxy = self.get_box_for_xy_bounds(
        )

        chip_bounds_xy = (chip_minx, chip_miny, chip_maxx, chip_maxy)

        if box_plus_buffer:
            # Based on component selection, determine the bounds for box_plus_buffer.
            frames = None
            frames_with_jj = None
            path_and_poly_with_valid_comps = None

            if case == 2:  # One or more components not in QDesign.
                self.design.logger.warning("One or more components not found.")
            elif case == 1:  # Render all components
                frames = [path_dataframe, poly_dataframe]
                frames_with_jj = [
                    path_dataframe, poly_dataframe, junction_dataframe
                ]
            else:  # Strict subset rendered.
                mask_path = path_dataframe['component'].isin(qcomp_ids)
                mask_poly = poly_dataframe['component'].isin(qcomp_ids)
                mask_junction = junction_dataframe['component'].isin(qcomp_ids)
                subset_path_df = path_dataframe[mask_path]
                subset_poly_df = poly_dataframe[mask_poly]
                subset_junction_df = junction_dataframe[mask_junction]
                frames = [subset_path_df, subset_poly_df]
                frames_with_jj = [
                    subset_path_df, subset_poly_df, subset_junction_df
                ]

            #Concat the frames and then determine the total bounds of all the geometries.
            # maybe, change name to package_cavity
            path_and_poly_with_valid_comps = pd.concat(frames,
                                                       ignore_index=True)
            path_poly_and_junction_valid_comps = pd.concat(frames_with_jj,
                                                           ignore_index=True)
            minx, miny, maxx, maxy = list(
                path_and_poly_with_valid_comps['geometry'].total_bounds)
            # minx, miny, maxx, maxy = list(
            #     pd.concat(frames, ignore_index=True)['geometry'].total_bounds)
            # # Add the buffer, using options for renderer.
            # x_buff = parse_entry(self._options["x_buffer_width_mm"])
            # y_buff = parse_entry(self._options["y_buffer_width_mm"])

            minx -= x_buff
            miny -= y_buff
            maxx += x_buff
            maxy += y_buff
            box_for_xy_bounds = (minx, miny, maxx, maxy)

            safe_xy_bounds = self.ensure_component_box_smaller_than_chip_box_(
                box_for_xy_bounds, chip_bounds_xy)

            return safe_xy_bounds, path_and_poly_with_valid_comps, path_poly_and_junction_valid_comps, self.chip_names_matched, self.valid_chip_names
        else:  # Incorporate all the chip sizes.

            frames = [path_dataframe, poly_dataframe]
            frames_with_jj = [
                path_dataframe, poly_dataframe, junction_dataframe
            ]
            path_and_poly_with_valid_comps = pd.concat(frames,
                                                       ignore_index=True)
            path_poly_and_junction_valid_comps = pd.concat(frames_with_jj,
                                                           ignore_index=True)
            return chip_bounds_xy, path_and_poly_with_valid_comps, path_poly_and_junction_valid_comps, self.chip_names_matched, self.valid_chip_names

    @classmethod
    def ensure_component_box_smaller_than_chip_box_(
            cls, box_for_xy_bounds: Tuple, chip_bounds_xy: Tuple) -> Tuple:
        """If the box_plus_buffer is larger than the aggregate chip bounds from DesignPlanar,
        use the chip bounds as the cutoff.

        Args:
            box_for_xy_bounds (Tuple): In xy plane, the bounding box for the components to render.
                                    Box from QGeometry tables.
            chip_bounds_xy (Tuple): In xy plane, the bounding box for aggregate chip size.
                                    Box from MultiPlanar chip size.

        Returns:
            Tuple: In xy plane, the box_for_xy_bounds will be limited by the chip_bounds_xy
                                    if box_for_xy_bounds is larger than chip_bounds_xy.
                                    If chip_bounds_xy is None (bad input), no checking
                                    will happen and box_for_xy_bounds will be returned.
        """

        chip_minx, chip_miny, chip_maxx, chip_maxy = chip_bounds_xy
        if chip_minx is None or chip_miny is None or chip_maxx is None or chip_maxy is None:
            input_xy_box = deepcopy(box_for_xy_bounds)
            return input_xy_box

        box_minx, box_miny, box_maxx, box_maxy = box_for_xy_bounds
        safe_xy_box = list()
        #  Keep the order of appends in this way.  It should match (minx, miny, maxx, maxy)

        #yapf: disable
        safe_xy_box.append(chip_minx) if box_minx < chip_minx else safe_xy_box.append(box_minx)
        safe_xy_box.append(chip_miny) if box_miny < chip_miny else safe_xy_box.append(box_miny)
        safe_xy_box.append(chip_maxx) if box_maxx > chip_maxx else safe_xy_box.append(box_maxx)
        safe_xy_box.append(chip_maxy) if box_maxy > chip_maxy else safe_xy_box.append(box_maxy)
        #yapf: enable

        return tuple(safe_xy_box)

    def get_box_for_xy_bounds(
            self
    ) -> Union[None, Union[Tuple[float, float, float, float], None]]:
        """Assuming the chip size is used from Multiplanar design, and list of chip_names
        comes from layer_stack that will be used to determine the box size for simulation.

        Returns:
            Union[None, Union[Tuple[float, float, float, float], None]]:
                                None if not able to get the chip information
                                Tuple holds the box to use for simulation [minx, miny, maxx, maxy]
        """

        self.chip_names_matched, self.valid_chip_names = self.are_all_chipnames_in_design(
        )

        minx, miny, maxx, maxy = None, None, None, None
        if self.chip_names_matched:
            # Using the chip size from Multiplanar design and z_coord from layer_stack, get the box
            # for chip_name in self.chip_names_matched:
            for chip_name in self.valid_chip_names:
                if self.design.chips[chip_name]['size']:
                    chip_box, return_code = self.get_x_y_for_chip(chip_name)

                    if return_code == 0:  # All was found and good.
                        minx, miny, maxx, maxy = determine_larger_box(
                            minx, miny, maxx, maxy, chip_box)

                else:
                    self.chip_size_not_in_chipname_within_design(chip_name)

        return minx, miny, maxx, maxy

    def are_all_chipnames_in_design(self) -> Tuple[bool, Union[set, None]]:
        """Using chip names in layer_stack information,
        then check if the information is in MultiPlanar design.

        Returns:
            Tuple[bool, Union[set, None]]: bool if there is a key in design with same chip_name as in layer_stack
                                        Union has either None if the names don't match,
                                        or the set of chip_names that can be used.
        """

        chip_set_from_design = set(self.design.chips.keys())
        chip_set_from_layer_stack = self.design.ls.get_unique_chip_names()
        if not chip_set_from_layer_stack.issubset(chip_set_from_design):
            self.chip_names_not_in_design(chip_set_from_layer_stack,
                                          chip_set_from_design)
            return False, None

        return True, chip_set_from_layer_stack

    def get_x_y_for_chip(self, chip_name: str) -> Tuple[tuple, int]:
        """If the chip_name is in self.chips, along with entry for size
        information then return a tuple=(minx, miny, maxx, maxy). Used for
        subtraction while exporting design.

        Args:
            chip_name (str): Name of chip that you want the size of.

        Returns:
            Tuple[tuple, int]:
            tuple: The exact placement on rectangle coordinate (minx, miny, maxx, maxy).
            int: 0=all is good
            1=chip_name not in self._chips
            2=size information missing or no good
        """
        x_y_location = tuple()

        if chip_name in self.design.chips:
            if 'size' in self.design.chips[chip_name]:

                size = self.design.parse_value(
                    self.design.chips[chip_name]['size'])
                if      'center_x' in size               \
                    and 'center_y' in size          \
                    and 'size_x' in size            \
                    and 'size_y' in size:
                    if type(size.center_x) in [int, float] and \
                            type(size.center_y) in [int, float] and \
                            type(size.size_x) in [int, float] and \
                            type(size.size_y) in [int, float]:
                        x_y_location = (
                            size['center_x'] - (size['size_x'] / 2.0),
                            size['center_y'] - (size['size_y'] / 2.0),
                            size['center_x'] + (size['size_x'] / 2.0),
                            size['center_y'] + (size['size_y'] / 2.0))
                        return x_y_location, 0

                    self.design.logger.warning(
                        f'Size information within self.design.chips[{chip_name}][\"size\"]'
                        f' is NOT an int or float.')
                    return x_y_location, 2

                self.design.logger.warning(
                    'center_x or center_y or size_x or size_y '
                    f' NOT in self.design.chips[{chip_name}][\"size\"]')
                return x_y_location, 2

            self.design.logger.warning(
                f'Information for size in NOT in self.design.chips[{chip_name}]'
                ' dict. Return "None" in tuple.')
            return x_y_location, 2

        self.design.logger.warning(
            f'Chip name "{chip_name}" is not in self.design.chips dict. Return "None" in tuple.'
        )
        return x_y_location, 1


######### Warnings and Errors##################################################

    def chip_names_not_in_design(self, layer_stack_names: set,
                                 design_names: set):
        """
        Tell user to check the chip name and data in design.

        Args:
            layer_stack_names (set): Chip names from layer_stack.
            design_names (set): Chip names from design.
        """
        self.design.logger.warning(
            f'\nThe chip_names from layer_stack are not in design. '
            f'\n The names in layer_stack:{layer_stack_names}.'
            f'\n The names in design:{design_names}.')

    def chip_size_not_in_chipname_within_design(self, chip_name: str):
        """
        Tell user to check the chip size data within design.

        Args:
            chip_name (str): Chip names from layer_stack.
        """
        self.design.logger.error(
            f'\nThe chip_name:{chip_name} within design. '
            f'\n Update your QDesign or subclass to see confirm the size information is provided.'
        )



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Logging widget.

Credits, based on:
    https://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt
    and thanks to Phil Reinhold
"""
import collections
import html
import logging
import random
from pathlib import Path

from PySide6 import QtGui
from PySide6.QtCore import Qt
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QDockWidget, QTextEdit

from .... import Dict, __version__, config
from ...utility._handle_qt_messages import slot_catch_error

from ... import config
if not config.is_building_docs():
    from ....toolbox_python.utility_functions import clean_name, monkey_patch

__all__ = ['QTextEditLogger', 'LogHandler_for_QTextLog']


class QTextEditLogger(QTextEdit):
    """A text edit logger class.

    This class extends the `QTextEdit` class
    """
    timestamp_len = 19
    _logo = 'metal_logo.png'

    def __init__(self, img_path='/', dock_window: QDockWidget = None):
        """Widget to handle logging. Based on QTextEdit, an advanced WYSIWYG
        viewer/editor supporting rich text formatting using HTML-style tags. It
        is optimized to handle large documents and to respond quickly to user
        input.

        Get as:
            gui.ui.log_text

        Args:
            img_path (str): Path to images.  Defaults to '/'.
            dock_window (QDockWidget): The dock window widget.  Defaults to None.
        """
        super().__init__()

        self.img_path = img_path
        self.dock_window = dock_window

        # handles the loggers
        # dict for what loggers we track and if we should show or not
        self.tracked_loggers = Dict()
        self.handlers = Dict()
        self._actions = Dict()  # menu actions. Must be an ordered Dict!
        self._auto_scroll = True  # autoscroll to end or not
        self._show_timestamps = False
        self._level_name = ''

        # Props of the Widget
        self.setTextInteractionFlags(Qt.TextSelectableByMouse |
                                     Qt.TextSelectableByKeyboard)
        self.text_format = QtGui.QTextCharFormat()
        self.text_format.setFontFamily('Consolas')

        self.logged_lines = collections.deque(
            [], config.GUI_CONFIG.logger.num_lines)

        self.setup_menu()

    def toggle_autoscroll(self, checked: bool):
        """Toggle the autoscroll.

        Args:
            checked (bool): True to toggle on, False otherwise
        """
        self._auto_scroll = bool(checked)

    def toggle_timestamps(self, checked: bool):
        """Toggle the timestamp.

        Args:
            checked (bool): True to toggle on, False otherwise
        """
        self._show_timestamps = bool(checked)
        self.show_all_messages()

    def setup_menu(self):
        """Setup the menu."""
        # Behaviour for menu: the widget displays its QWidget::actions() as context menu.
        # i.e., a local context menu
        self.setContextMenuPolicy(Qt.ActionsContextMenu)

        ###################
        # Click actions
        actions = self._actions

        actions.clear_log = QAction('&Clear log', self, triggered=self.clear)
        actions.print_tips = QAction('&Show tips ',
                                     self,
                                     triggered=self.print_all_tips)

        actions.separator = QAction(self)
        actions.separator.setSeparator(True)

        ###################
        # Toggle actions
        self.action_scroll_auto = QAction('&Autoscroll',
                                          self,
                                          checkable=True,
                                          checked=True)
        self.action_scroll_auto.toggled.connect(self.toggle_autoscroll)

        self.action_show_times = QAction('&Show timestamps',
                                         self,
                                         checkable=True,
                                         checked=False)
        self.action_show_times.toggled.connect(self.toggle_timestamps)

        ###################
        # Filter level actions
        def make_trg(lvl):
            """Make a trg.

            Args:
                lvl (logging.level): The level of logging, eg.., logging.ERROR

            Returns:
                str: Value of the name attribute
            """
            name = f'set_level_{lvl}'
            # self.set_level(lvl)
            setattr(self, name, lambda: self.set_level(lvl))
            func = getattr(self, name)
            return func

        actions.debug = QAction('Set filter level:  Debug',
                                self,
                                triggered=make_trg(logging.DEBUG))
        actions.info = QAction('Set filter level:  Info',
                               self,
                               triggered=make_trg(logging.INFO))
        actions.warning = QAction('Set filter level:  Warning',
                                  self,
                                  triggered=make_trg(logging.WARNING))
        actions.error = QAction('Set filter level:  Error',
                                self,
                                triggered=make_trg(logging.ERROR))

        actions.separator2 = QAction(self)
        actions.separator2.setSeparator(True)

        actions.loggers = QAction('Show/hide messages for logger:',
                                  self,
                                  enabled=False)

        # Add actions to actin context menu
        self.addActions(list(actions.values()))

    def welcome_message(self):
        """Display the welcome message."""
        img_txt = ''

        # Logo
        img_path = Path(self.img_path) / self._logo
        if img_path.is_file():
            img_txt = f'<img src="{img_path}" height=80>'
        else:
            print(
                'WARNING: welcome_message could not locate img_path={img_path}')

        # Main message
        text = f'''<span class="INFO">{' '*self.timestamp_len}
        <br>
        <table border="0" width="100%" ID="tableLogo" style="margin: 0px;">
            <tr>
                <td align="center">
                    <h3 align="center"  style="text-align: center;">
                        Welcome to Qiskit Metal!
                    </h3>
                    v{__version__}
                </td>
                <td>  {img_txt} </td>
            </tr>
        </table>
        </span>
        <b>Tip: </b> {random.choice(config.GUI_CONFIG['tips'])}'''
        self.log_message(text, format_as_html=2)

    def print_all_tips(self):
        """Prints all available tips in the log window."""
        for tip in config.GUI_CONFIG['tips']:
            self.log_message(
                f'''<br><span class="INFO">{' '*self.timestamp_len} \u2022 {tip} </span>'''
            )

    def set_level(self, level: int):
        """Set level on all handlers.

        Args:
            level (logging.level): The level of logging, eg.., logging.ERROR
        """
        print(f'Setting level: {level}')
        self.set_window_title_level(level)
        for name, handler in self.handlers.items():
            handler.setLevel(level)

    def set_window_title_level(self, level: int):
        """Set the window title level.

        Args:
            level (int): the level
        """
        self._level_name = logging.getLevelName(level).lower()
        if self._level_name not in ['']:
            self.dock_window.setWindowTitle(
                f'Log  (filter >= {self._level_name})')
        else:
            self.dock_window.setWindowTitle(
                f'Log (right click log for options)')
        return self._level_name

    def add_logger(self, name: str, handler: logging.Handler):
        """Adds a logger to the widget.

            - adds `true bool`   to self.tracked_loggers for on/off to show
            - adds `the handler` in self.handlers
            - adds an action to the menu for the self.traceked_loggers

        For example, a logger handler is added with
            `gui.logger.addHandler(self._log_handler)`
        where
            `_log_handler is LogHandler_for_QTextLog`

        Args:
            name (string): Name of logger to be added
            handler (logging.Handler): Handler
        """
        if name in self.tracked_loggers:
            return

        self.tracked_loggers[name] = True
        # can this be a problem if handler has multiple names?
        self.handlers[name] = handler

        #############################################
        # Monkey patch add function
        func_name = f'toggle_{clean_name(name)}'

        def toggle_show_log(self2, val: bool):
            """Toggle the value of the.

            Args:
                self2 (QTextEdit): self
                val (bool): True or False

            Example:
                <bound method QTextEditLogger.add_logger.<locals>.toggle_show_log
                of <qiskit_metal._gui.widgets.log_widget.log_metal.QTextEditLogger object at 0x7fce3a500c18>>
            """
            self2.tracked_loggers[name] = bool(val)

        monkey_patch(self, toggle_show_log, func_name=func_name)
        #############################################

        # Add action
        action = QAction(f' - {name}',
                         self,
                         checkable=True,
                         checked=True,
                         triggered=getattr(self, func_name))
        self._actions[f'logger_{name}'] = action
        self.addAction(action)

        # style TODO: move
        self.document().setDefaultStyleSheet(config.GUI_CONFIG.logger.style)

        # is this the first logger we added
        if len(self.tracked_loggers) == 1:
            self.set_window_title_level(handler.level)

    def get_all_checked(self):
        """Get all the checked items.

        Returns:
            list: List of checked items
        """
        res = []
        for name, isChecked in self.tracked_loggers.items():
            if isChecked:
                res += [name]
        return res

    def show_all_messages(self):
        """Clear and reprint all log lines, thus refreshing toggles for
        timestamp, etc."""
        self.clear()
        for name, record in self.logged_lines:
            if name in self.get_all_checked():
                self.log_message(record, name != 'Errors')

    def log_message_to(self, name, record):
        """Set where to log messages to.

        Args:
            name (str): The name
            record (bool): True to send to records, False otherwise
        """
        self.logged_lines.append((name, record))
        if name in self.get_all_checked():
            self.log_message(record, name != 'Errors')

    def log_message(self, message, format_as_html=True):
        """Do the actual logging.

        Args:
            message (str): The message to log.
            format_as_html (bool): True to format as HTML, False otherwise.  Defaults to True.
        """
        # set the write positon
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.End)
        cursor.insertBlock()  # add a new block, which makes a new line

        # add message
        if format_as_html == True:  # pylint: disable=singleton-comparison

            if not self.action_show_times.isChecked():
                # remove the timestamp -- assumes that this has been formatted
                # with pre tag by LogHandler_for_QTextLog
                res = message.split('<pre>', 1)
                if len(res) == 2:
                    message = res[0] + '<pre>' + res[1][1 + self.timestamp_len:]
                else:
                    pass  # print(f'Warning incorrect: {message}')

            cursor.insertHtml(message)

        elif format_as_html == 2:
            cursor.insertHtml(message)

        else:
            cursor.insertText(message, self.text_format)

        # make sure that the message is visible and scrolled ot
        if self.action_scroll_auto.isChecked():
            self.moveCursor(QtGui.QTextCursor.End)
            self.moveCursor(QtGui.QTextCursor.StartOfLine)
        self.ensureCursorVisible()

    def remove_handlers(self, logger):
        """Call on clsoe window to remove handlers from the logger."""
        for name, handler in self.handlers.items():
            if handler in logger.handlers:
                logger.handlers.remove(handler)


class LogHandler_for_QTextLog(logging.Handler):
    """Class to handle GUI logging. Handler instances dispatch logging events
    to specific destinations.

    For formatting:
        https://docs.python.org/3/library/logging.html#logrecord-attributes
        _log_string = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')     # create formatter and add it to the handlers

    This class extends the `logging.Handler` class.
    """

    def __init__(self,
                 name,
                 parent,
                 log_qtextedit: QTextEditLogger,
                 logger: logging.Logger,
                 log_string=None):
        """
        Args:
            name (str): The name.
            parent (logger): The parent logger.
            log_qtextedit (QTextEditLogger): Text edit logger.
            logger (logging.Logger): The logger.
            log_string (str): The log string.  Defaults to None.
        """
        super().__init__()

        self.name = name
        self.log_qtextedit = log_qtextedit
        self.setLevel(int(logger.level))
        self._logger = logger  # not sure if good idea to do this

        # Formatter
        if isinstance(log_string, str):
            self._log_string = log_string
        else:
            self._log_string = f'%(asctime).{QTextEditLogger.timestamp_len}s %(name)s: %(message)s [%(module)s.%(funcName)s]'
        self._log_formatter = logging.Formatter(self._log_string)
        self.setFormatter(self._log_formatter)

        # Add formatter to
        self.log_qtextedit.add_logger(name, self)

        self._logger.addHandler(self)  # ADD HANDLER!

    def emit(self, record):
        """Do whatever it takes to actually log the specified logging record.
        Converts the characters '&', '<' and '>' in string s to HTML-safe
        sequences. Used to display text that might contain such characters in
        HTML.

        Args:
            record (LogRecord): The log recorder
        """
        # print(record)
        #self.log_qtextedit.record = record
        #self.log_qtextedit.zz = self

        html_record = html.escape(self.format(record))
        html_log_message = '<span class="%s"><pre>%s</pre></span>' % (
            record.levelname, html_record)
        try:
            self.log_qtextedit.log_message_to(self.name, html_log_message)
        except RuntimeError as e:
            # trying to catch
            #  RuntimeError('wrapped C/C++ object of type QTextEditLogger has been deleted',)
            print(f'Logger issue: {e}')
            self._logger.handlers.remove(self)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""File contains some config definitions.

Mostly internal.
"""

import logging, collections
from typing import List

__all__ = ['setup_logger', 'LogStore']


def setup_logger(logger_name,
                 log_format,
                 log_datefmt,
                 level_stream=logging.INFO,
                 level_base=logging.DEBUG,
                 force_set=False,
                 capture_warnings=None,
                 propagate=False,
                 create_stream=True) -> logging.Logger:
    """Setup the logger to work with jupyter and command line.

    `level_stream` and `level_base`:
    You can set a different logging level for each logging handler, however you have
    to set the logger's level to the "lowest".

    Integrates logging with the warnings module.

    Args:
        logger_name (str): Name of the log.
        log_format (format): Format of the log.
        log_datefmt (format): Format of the date.
        level_stream (log level): Log level of the stream.  Defaults to logging.INFO.
        level_base (log level): Log level of the base.  Defaults to logging.DEBUG.
        force_set (bool): True to force.  Defaults to False.
        capture_warnings (bool): True to capture warnings.  Defaults to None.
        propagate (bool): True to propagate.  Defaults to False.
        create_stream (bool): True to create the stream.  Defaults to True.

    Returns:
        logging.Logger: The logger

    To see the logging levels you can use:

    .. code-block:: python

        print(logger)
        print(logger.zkm_c_handler)
        print(gui._log_handler)
    """

    logger = logging.getLogger(logger_name)  # singleton

    if force_set or not len(logger.handlers):

        logger.setLevel(level_base)

        # Used to integrate logging with the warnings module.
        # Warnings issued by the warnings module will be redirected to the logging system.
        # Specifically, a warning will be formatted using warnings.formatwarning() and the resulting
        # string logged to a logger named 'py.warnings' with a severity of WARNING.
        if capture_warnings is not None:
            logging.captureWarnings(capture_warnings)

        # Jupyter notebooks already has a stream handler on the default log.
        # Do not propage upstream to the root logger.
        # https://stackoverflow.com/questions/31403679/python-logging-module-duplicated-console-output-ipython-notebook-qtconsole
        logger.propagate = propagate

        if create_stream:
            # Sends logging output to streams such as sys.stdout,
            # sys.stderr or any file-like object
            c_handler = logging.StreamHandler()

            # Format. Unlike the root logger, a custom logger can't be configured
            # using basicConfig().
            c_format = logging.Formatter(log_format, datefmt=log_datefmt)
            c_handler.setFormatter(c_format)

            # Add Hanlder with format and set level
            logger.addHandler(c_handler)
            c_handler.setLevel(level_stream)

            # save references for ease
            logger.zkm_c_handler = c_handler
            logger.zkm_c_format = c_format

    return logger


class LogStore(collections.deque):
    """Wrapper over collections.deque that ensures most recently added items
    (which should be strings) are at the "front" of the queue (i.e. left of the
    array).

    Each QDesign instantiation has a LogStore object used to keep track
    of logs for the Build History display.
    """

    def __init__(self,
                 title: str,
                 log_limit: int,
                 _previous_builds: List[str] = [],
                 *args,
                 **kwargs):
        super().__init__(maxlen=log_limit, *args, **kwargs)
        self._title = title
        for i in _previous_builds:
            self.appendleft(i)
        self._next_available_index = 0

    def data(self):
        #returns COPY of data
        return list(self)

    def add(self, log: str):
        self.appendleft(log)

    def add_success(self, log: str):
        self.appendleft(log)

    def add_error(self, log: str):
        self.appendleft(
            "ERROR " + log
        )  #used to differentiate error strings from success strings where logs are displayed

    @property
    def title(self):
        """The title is an easy way to differentiate between multiple instances
        of LogStore, each with different data."""
        return self._title



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import heapq
import numpy as np
from qiskit_metal import Dict
from qiskit_metal.qlibrary.core import QRoutePoint
from .anchored_path import RouteAnchors
from qiskit_metal.toolbox_metal import math_and_overrides as mao
from collections import OrderedDict
from qiskit_metal.toolbox_metal.exceptions import QiskitMetalDesignError


class RoutePathfinder(RouteAnchors):
    """
    Non-meandered CPW class that combines A* pathfinding algorithm with
    simple 1-, 2-, or S-shaped segment checks and user-specified anchor points.
    1. A* heap modified to prioritize paths with shortest length_travelled + Manhattan distance to destination.
    2. Checks if connect_simple is valid each time we pop from the heap. If so, use it, otherwise proceed with A*.
    3. Tweaks connect_simple to account for end anchor direction in determining which CPW (elbow or S-segment) to use.

    .. meta::
        :description: Route Pathfinder

    RouteAnchors Default Options:
        * anchors: OrderedDict -- Intermediate anchors only; doesn't include endpoints
        * advanced: Dict
            * avoid_collision: 'false' -- true/false, defines if the route needs to avoid collisions

    Default Options:
        * step_size: '0.25mm' -- Length of the step for the A* pathfinding algorithm
        * advanced: Dict
            * avoid_collision: 'true' -- true/false, defines if the route needs to avoid collisions
    """

    default_options = Dict(step_size='0.25mm',
                           advanced=Dict(avoid_collision='true'))
    """Default options"""

    TOOLTIP = """ Non-meandered CPW class that combines A* pathfinding algorithm with
    simple 1-, 2-, or S-shaped segment checks and user-specified anchor points."""

    def connect_astar_or_simple(self, start_pt: QRoutePoint,
                                end_pt: QRoutePoint) -> list:
        """Connect start and end via A* algo if connect_simple doesn't work.

        Args:
            start_direction (np.array): Vector indicating direction of starting point
            start (np.array): 2-D coordinates of first anchor
            end (np.array): 2-D coordinates of second anchor

        Returns:
            List of vertices of a CPW going from start to end

        Raises:
            QiskitMetalDesignError: If the connect_simple() has failed.
        """

        start_direction = start_pt.direction
        start = start_pt.position
        end_direction = end_pt.direction
        end = end_pt.position

        step_size = self.parse_options().step_size

        starting_dist = sum(
            abs(end - start))  # Manhattan distance between start and end
        key_starting_point = (starting_dist, start[0], start[1])
        pathmapper = {key_starting_point: [starting_dist, [start]]}
        # pathmapper maps tuple(total length of the path from self.start + Manhattan distance to destination, coordx, coordy) to [total length of
        # path from self.start, path]
        visited = set(
        )  # maintain record of points we've already visited to avoid self-intersections
        visited.add(tuple(start))
        # TODO: add to visited all of the current points in the route, to prevent self intersecting
        priority_queue = list()  # A* priority queue. Implemented as heap
        priority_queue.append(key_starting_point)
        # Elements in the heap are ordered by the following:
        # 1. The total length of the path from self.start + Manhattan distance to destination
        # 2. The x coordinate of the latest point
        # 3. The y coordinate of the latest point

        while priority_queue:
            tot_dist, x, y = heapq.heappop(
                priority_queue
            )  # tot_dist is the total length of the path from self.start + Manhattan distance to destination
            length_travelled, current_path = pathmapper[(tot_dist, x, y)]
            # Look in forward, left, and right directions a fixed distance away.
            # If the line segment connecting the current point and this next one does
            # not collide with any bounding boxes in design.components, add it to the
            # list of neighbors.
            neighbors = list()
            if len(current_path) == 1:
                # At starting point -> initial direction is start direction
                direction = start_direction
            else:
                # Beyond starting point -> look at vector difference of last 2 points along path
                direction = current_path[-1] - current_path[-2]
            # The dot product between direction and the vector connecting the current
            # point and a potential neighbor must be non-negative to avoid retracing.

            # Check if connect_simple works at each iteration of A*
            try:
                simple_path = self.connect_simple(
                    QRoutePoint(np.array([x, y]), direction),
                    QRoutePoint(end, end_direction))
            except QiskitMetalDesignError:
                simple_path = None
                # try the pathfinder algorithm
                pass

            if simple_path is not None:
                current_path.extend(simple_path)
                return current_path

            for disp in [
                    np.array([0, 1]),
                    np.array([0, -1]),
                    np.array([1, 0]),
                    np.array([-1, 0])
            ]:
                # Unit displacement in 4 cardinal directions
                if mao.dot(disp, direction) >= 0:
                    # Ignore backward direction
                    curpt = current_path[-1]
                    nextpt = curpt + step_size * disp
                    if self.unobstructed([curpt, nextpt]):
                        neighbors.append(nextpt)
            for neighbor in neighbors:
                if tuple(neighbor) not in visited:
                    new_remaining_dist = sum(abs(end - neighbor))
                    new_length_travelled = length_travelled + step_size
                    new_path = current_path + [neighbor]
                    if new_remaining_dist < 10**-8:
                        # Destination has been reached within acceptable error tolerance (errors due to rounding in Python)
                        return new_path[:-1] + [
                            end
                        ]  # Replace last element of new_path with end since they're basically the same
                    heapq.heappush(priority_queue,
                                   (new_length_travelled + new_remaining_dist,
                                    neighbor[0], neighbor[1]))
                    pathmapper[(new_length_travelled + new_remaining_dist,
                                neighbor[0], neighbor[1])] = [
                                    new_length_travelled, new_path
                                ]
                    visited.add(tuple(neighbor))
        return [
        ]  # Shouldn't actually reach here - if it fails, there's a convergence issue

    def make(self):
        """Generates path from start pin to end pin."""
        p = self.parse_options()
        anchors = p.anchors

        # Set the CPW pins and add the points/directions to the lead-in/out arrays
        self.set_pin("start")
        self.set_pin("end")

        # Align the lead-in/out to the input options set from the user
        start_point = self.set_lead("start")
        end_point = self.set_lead("end")

        self.intermediate_pts = OrderedDict()
        for arc_num, coord in anchors.items():
            arc_pts = self.connect_astar_or_simple(self.get_tip(),
                                                   QRoutePoint(coord))
            if arc_pts is None:
                self.intermediate_pts[arc_num] = [coord]
            else:
                self.intermediate_pts[arc_num] = np.concatenate(
                    [arc_pts, [coord]], axis=0)
        arc_pts = self.connect_astar_or_simple(self.get_tip(), end_point)
        if arc_pts is not None:
            self.intermediate_pts[len(anchors)] = np.array(arc_pts)

        # concatenate all points, transforming the dictionary into a single numpy array
        self.trim_pts()
        self.intermediate_pts = np.concatenate(list(
            self.intermediate_pts.values()),
                                               axis=0)

        # Make points into elements
        self.make_elements(self.get_points())



## -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=relative-beyond-top-level
# pylint: disable-msg=import-error
# pylint: disable-msg=import-self
"""
=================================================
Toolbox_metal (:mod:`qiskit_metal.toolbox_metal`)
=================================================

.. currentmodule:: qiskit_metal.toolbox_metal

This folder contains code that is more specific to Metal.

Custom Exceptions
-----------------

.. autosummary::
    :toctree: .

    QiskitMetalExceptions
    QiskitMetalDesignError


Submodules
---------------

.. autosummary::
    :toctree: .

    about
    import_export
    math_and_overrides
    parsing
    layer_stack_handler
    bounds_for_path_and_poly_tables
    determine_larger_box

"""

from .parsing import parse_value
from .parsing import is_variable_name
from .parsing import is_numeric_possible
from .parsing import parse_units
from .layer_stack_handler import LayerStackHandler
from .bounds_for_path_and_poly_tables import BoundsForPathAndPolyTables, determine_larger_box

from .. import config
if config.is_building_docs():
    from . import about
    from .exceptions import QiskitMetalDesignError
    from .exceptions import QiskitMetalExceptions
    from . import import_export
    from . import parsing
    from . import math_and_overrides
    from . import layer_stack_handler
    from . import bounds_for_path_and_poly_tables



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6 import QtCore
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtGui import QFont

from .... import config


class PropValTable(QAbstractTableModel):
    """Design variables table model that shows variable name and dimension,
    both with and without units.

    Thie class extends the `QAbstractTableModel` class.

    Access:
        gui.variables_window.model
    """

    __refreshtime = 500  # 0.5 second refresh time

    def __init__(self, design=None, gui=None, view: 'RightClickView' = None):
        """
        Args:
            design (QDesign): The QDesign.  Defaults to None.
            gui (MetalGUI): The MetalGUI.  Defaults to None.
            view (view): The view.  Defaults to None.
        """
        super().__init__()
        self._design = design
        self._gui = gui
        self._view = view
        self._row_count = -1
        self._start_timer()

    def set_design(self, design):
        """Set the design.

        Args:
            design (QDesign): The design
        """
        self._design = design
        # self.modelReset.emit()
        self.auto_refresh()
        # refresh table or something if needed

    @property
    def design(self):
        """Returns the design."""
        return self._design

    @property
    def _data(self) -> dict:
        """Returns the data."""
        if self._design:
            return self._design.variables

    def _start_timer(self):
        """Start and continuously refresh timer in background to keep the total
        number of rows up to date."""
        self.timer = QtCore.QTimer(self)
        self.timer.start(self.__refreshtime)
        self.timer.timeout.connect(self.auto_refresh)

    def auto_refresh(self):
        """Do an automatic refresh."""
        new_row_count = self.rowCount(self)
        if self._row_count != new_row_count:
            # Wrap the reset logic in beginResetModel and endResetModel
            self.beginResetModel()
            try:
                # self.modelReset.emit()
                self._row_count = new_row_count
            finally:
                self.endResetModel()
            if self._view:
                self._view.resizeColumnsToContents()

    def rowCount(self, index: QModelIndex) -> int:
        """Count the number of rows.

        Args:
            index (QModelIndex): Not used

        Returns:
            int: Ihe number of rows
        """
        if self._design:
            return len(self._data)
        else:
            return 0

    def columnCount(self, index: QModelIndex) -> int:
        """Count the number of columns.

        Args:
            index (QModelIndex): Not used

        Returns:
            int: Yhe number of columns.  Always returns 3
        """
        return 3

    def data(self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole):
        """Return data for corresponding index and role.

        Args:
            index (QModelIndex): The index of the data
            role (QT.ItemDataRole): The data role.  Defaults to Qt.DisplayRole.

        Returns:
            ojbect: The data
        """
        self._index = index
        row = index.row()
        column = index.column()

        if role == Qt.DisplayRole:
            if column == 0:
                return str(list(self._data.keys())[row])
            elif column == 1:
                return str(self._data[list(self._data.keys())[row]])
            elif column == 2:
                return str(
                    self.design.parse_value(self._data[list(
                        self._data.keys())[row]]))

        # double clicking
        elif role == Qt.EditRole:
            return self.data(index, Qt.DisplayRole)

        elif (role == Qt.FontRole) and (column == 0):
            font = QFont()
            font.setBold(True)
            return font

    def setData(self,
                index: QModelIndex,
                value: str,
                role: Qt.ItemDataRole = Qt.EditRole) -> bool:
        """Modify either key or value (Property or Value) of dictionary
        depending on what the user selected manually on the table.

        Args:
            index (QModelIndex): The index
            value (str): The data value
            role (Qt.ItemDataRole): The role of the data.  Defaults to Qt.EditRole.

        Returns:
            bool: True if successful; otherwise returns False.
        """
        r = index.row()
        c = index.column()

        if value:

            if c == 0:
                # TODO: LRU Cache for speed?
                oldkey = list(self._data.keys())[r]
                if value != oldkey:
                    self.design.rename_variable(oldkey, value)
                    self._gui.rebuild()
                    return True

            elif c == 1:
                self._data[list(self._data.keys())[r]] = value
                self._gui.rebuild()
                return True

        return False

    def headerData(self,
                   section: int,
                   orientation: Qt.Orientation,
                   role: Qt.ItemDataRole = Qt.DisplayRole) -> str:
        """Get the headers to be displayed.

        Args:
            secion (int): Section number.
            orientation (Qt.Orientation): Orientation of the header.
            role (Qt.ItemDataRole): Role of the header.  Defaults to Qt.DisplayRole.

        Returns:
            str: The header
        """
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if section == 0:
                    return 'Variable name'
                elif section == 1:
                    return 'Value'
                else:
                    units = config.DefaultMetalOptions.default_generic.units
                    if self.design:
                        if hasattr(self.design, '_template_options'):
                            units = self.design.template_options.units
                    return f'Parsed value (in {units})'
            return str(section + 1)

    def removeRows(self, row: int, count: int = 1, parent=QModelIndex()):
        """Delete highlighted rows.

        Args:
            row (int): First row to delete.
            count (int): Number of rolws to delete.  Defaults to 1.
            parent (QModelIndex): Parent index.
        """
        self.beginRemoveRows(parent, row, row + count - 1)
        lst = list(self._data.keys())
        for k in range(row + count - 1, row - 1, -1):
            del self._data[lst[k]]
        self.endRemoveRows()

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """Determine how user may interact with each cell in the table.

        Args:
            index (QModelIndex): The index

        Returns:
            Qt.ItemFlags: The flags
        """
        if index.column() < 2:
            return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def add_row(self, key: str, val: str):
        """Add row with the given key/value.

        Args:
            key (str): The key
            val (str): The value
        """
        self._data[key] = val
        self._view.resizeColumnsToContents()



# -*- coding: utf-8 -*-

from re import search
import pandas as pd
from typing import List, Tuple, Dict
from typing import Union
from addict import Dict
import os
import logging

from .parsing import TRUE_STR, FALSE_STR


class LayerStackHandler():
    """Use DataFrame to hold information for multiple chips.

    For all the chips, do NOT reuse the layer-number.  However, within the dataframe,
    we can repeat layer numbers since a layer can have multiple datatypes.

    """

    Col_Names = [
        'chip_name', 'layer', 'datatype', 'material', 'thickness', 'z_coord',
        'fill'
    ]

    def __init__(self,
                 multi_planar_design: 'MultiPlanar',
                 fname: Union['str', None] = None) -> None:
        """Use the information in MultiPlanar design to parse_value
        and get the name of filename for layer stack.

        Args:
            multi_planar_design (MultiPlanar): Class that is child of QDesign.
        """

        self.multi_planar_design = multi_planar_design
        self.logger = self.multi_planar_design.logger  # type: logging.Logger

        self.filename_csv_df = None
        if hasattr(self.multi_planar_design,
                   'ls_filename') and self.multi_planar_design.ls_filename:
            #populate pandas table from data in this file.
            self.filename_csv_df = self.multi_planar_design.ls_filename
        elif fname:
            self.filename_csv_df = fname

        self.ls_df = None
        # self.column_names = [
        #     'chip_name', 'layer', 'datatype', 'material', 'thickness',
        #     'z_coord', 'fill'
        # ]
        self.layer_stack_default = Dict(chip_name=['main', 'main'],
                                        layer=[1, 3],
                                        datatype=[0, 0],
                                        material=['pec', 'silicon'],
                                        thickness=['2um', '-750um'],
                                        z_coord=['0um', '0um'],
                                        fill=['true', 'true'])

        self._init_dataframe()
        self.is_layer_data_unique()

    def _init_dataframe(self) -> None:
        """Must check if filename for layerstack is valid before trying to import to a pandas table.
        """

        if self.filename_csv_df:
            #populate pandas table from data in this file.
            abs_path = os.path.abspath(self.filename_csv_df)
            if os.path.isfile(abs_path):
                self._read_csv_df(abs_path)
            else:
                self.logger.error(
                    f'Not able to read file.'
                    f'File:{abs_path} not read. Check the name and path.')

        else:
            #enter very basic default data for pandas table.
            self.ls_df = pd.DataFrame(data=self.layer_stack_default)

    def get_layer_datatype_when_fill_is_true(self) -> Union[dict, None]:
        """Return layer/datatype rows where ``fill`` is True.

        Returns:
            dict | None: Mapping of (layer, datatype) -> dict of parsed column
            values, or None if no rows have ``fill=True``.
        """
        if self.ls_df is None:
            abs_path = os.path.abspath(self.filename_csv_df)
            self.logger.error(
                f'Not able to read file.'
                f'File:{abs_path} not read. Check the name and path.')

        mask = self.ls_df['fill'].astype(str).str.replace(
            "[']", "", regex=True).isin(TRUE_STR)
        search_result_df = self.ls_df[mask]

        if len(search_result_df) > 0:
            try:
                layer_datatype_fill = dict()
                for row in search_result_df.itertuples():
                    a_key = (row.layer, row.datatype)
                    layer_datatype_fill[a_key] = dict()
                    layer_datatype_fill[a_key]['layer'] = row.layer
                    layer_datatype_fill[a_key]['datatype'] = row.datatype
                    layer_datatype_fill[a_key][
                        'thickness'] = self.multi_planar_design.parse_value(
                            row.thickness.strip('\''))
                    layer_datatype_fill[a_key][
                        'z_coord'] = self.multi_planar_design.parse_value(
                            row.z_coord.strip('\''))
                    layer_datatype_fill[a_key]['material'] = row.material.strip(
                        '\'')
                    layer_datatype_fill[a_key][
                        'chip_name'] = row.chip_name.strip('\'')
                return layer_datatype_fill
            except Exception as ex:
                self.logger.error(
                    f'User is not using LayerStackHandler.get_layer_datatype_when_fill_is_true correctly. Check your input file.'
                    f'\nERROR:{ex}')
        return None

    def get_properties_for_layer_datatype(
            self,
            properties: List[str],
            layer_number: int,
            datatype: int = 0) -> Union[Tuple[Union[float, str, bool]], None]:
        """When user provides a layer and datatype, they can get properties
         from the layer_stack file. The allowed options for properties must
         be in Col_Names.  If any of the properties are not in Col_Names,
         None will be returned.  Otherwise a Tuple will be returned with properties
         in the same order as provided in input variable properties.

        Args:
            properties (List[str]): The column(s) within the layer stack that you want
            for a row based on layer, and datatype.
            layer_number (int): The layer number within the column denoted by layer.
            datatype (int, optional): The datatype within the column denoted
                                    by datatype. Defaults to 0.

        Returns:
            Union[Tuple[Union[float, str, bool]], None]: If the search data provided
                            in the arguments are not in the layer_stack file,
                            None will be returned.  If the search values are found in
                            layer_stack file, then a Tuple will be returned with the
                            requested properties in the same order as provided in
                            input variable denoted by properties.
        """
        if not properties:
            return None

        # Check if parameter is not a subset if Col_Names T.
        if not set(properties).issubset(set(self.Col_Names)):
            self._warning_properties(properties)
            return None

        props = Dict()
        thickness = 0.0
        z_coord = 0.0
        material = None
        fill_value = None
        chip_name = None

        if self.ls_df is None:
            abs_path = os.path.abspath(self.filename_csv_df)
            self.logger.error(
                f'Not able to read file.'
                f'File:{abs_path} not read. Check the name and path.')

        mask = (self.ls_df['layer'] == layer_number) & (self.ls_df['datatype']
                                                        == datatype)

        search_result_df = self.ls_df[mask]

        if len(search_result_df) > 0:
            try:
                thickness = self.multi_planar_design.parse_value(
                    search_result_df.thickness.iloc[0].strip('\''))
                z_coord = self.multi_planar_design.parse_value(
                    search_result_df.z_coord.iloc[0].strip('\''))
                material = search_result_df.material.iloc[0].strip('\'')
                chip_name = search_result_df.chip_name.iloc[0].strip('\'')
                value = search_result_df.fill.iloc[0].strip('\'')
                if value in TRUE_STR:
                    fill_value = True
                elif value in FALSE_STR:
                    fill_value = False
                else:
                    self.logger.warning(
                        f'The \"fill\" value is neither True nor False.'
                        f'You have:{value}.  '
                        f'Will return NULL for fill value.')
                props['thickness'] = thickness
                props['z_coord'] = z_coord
                props['material'] = material
                props['fill'] = fill_value
                props['chip_name'] = chip_name
            except Exception as ex:
                self._warning_search_minus_chip(layer_number, datatype, ex)
        result = list()
        for item in properties:
            result.append(props[item])
        return tuple(result)

    def is_layer_data_unique(self) -> bool:
        """For each layer number make sure the datatypes are unique.  A layers can
        #have multiple datatypes.

        Returns:
            bool: True if empty dataframe, True if for each layer, there is ONLY one datatype. Otherwise, False.
        """
        layer_nums = self.get_unique_layer_ints()
        if layer_nums:
            for num in layer_nums:
                mask = self.ls_df['layer'] == num
                search_result_num = self.ls_df[mask]
                if not search_result_num.datatype.is_unique:
                    self.logger.warning(
                        f'There WILL BE PROBLEMS since layer {num} does not have unique datatypes.'
                    )
                    return False

        return True

    def _read_csv_df(self, abs_path: str) -> None:
        #ASSUME that self.filename_csv_df is valid file path and name.
        #https://best-excel-tutorial.com/59-tips-and-tricks/485-how-to-display-a-single-quote-in-a-cell
        try:
            self.ls_df = pd.read_csv(abs_path)

        except BaseException as error:
            self.logger.warning(
                f'Not able to create pandas dataframe.'
                f'File:{abs_path} not imported. Expected a CSV formatted file.')
            self.logger.error(f'The exception: {error}')

    def get_unique_chip_names(self) -> set:
        """Get a set of unique names used in the layer_stack dataframe.

        Returns:
            set: Unique names used in the layer stack used as either default or provided by user.
        """
        names = self.ls_df['chip_name']
        result = set(names.str.strip('\''))
        return result

    def get_unique_layer_ints(self) -> set:
        """Get a set of unique layer ints used in the layer_stack dataframe.

        Returns:
            set: Unique layer numbers used in the layer stack used as either default or provided by user.
        """
        layers = self.ls_df['layer']
        return set(layers.unique())

    def _warning_properties(self, properties: list):
        """_Give warning if the properties is

        Args:
            parameters (list): _description_
        """
        self.logger.warning(
            f'The list for properties: {properties} is not a'
            f'subset of expected column names: {self.Col_Names}')

    def _warning_search(self, chip_name: str, layer_number: int, data_type: int,
                        ex: Exception):
        """Give warning when the layerstack pandas table doesn't
        have requested chip_name, layer, and datatype information.

        Args:
            chip_name (str): Name of chip which has the layer of interest.
            layer_number (int): The unique layer number through out all chips.
                            The same layer number can  not be used on multiple chips.
            datatype (int, optional): Datatype of the layer number.
        """

        self.logger.warning(
            f'\nERROR: {ex}'
            f'\nThere is an error searching in layer_stack dataframe using '
            f'chip_name={chip_name}, layer={layer_number}, datatype={data_type}'
        )

    def _warning_search_minus_chip(self, layer_number: int, data_type: int,
                                   ex: Exception):
        """Give warning when the layerstack pandas table doesn't
        have requested layer, and datatype information.

        Args:
            layer_number (int): The unique layer number through out all chips.
                            The same layer number can  not be used on multiple chips.
            datatype (int, optional): Datatype of the layer number.
        """

        self.logger.warning(
            f'\nERROR: {ex}'
            f'\nThere is an error searching in layer_stack dataframe using '
            f' layer={layer_number}, datatype={data_type}')

    def layer_stack_handler_pilot_error(self):
        """ The handler will return None if incorrect arguments are passed.
        """
        self.logger.error(
            f'User is not using LayerStackHandler.get_properties_for_chip_layer_datatype correctly.'
        )
